from pathlib import Path

import numpy as np
import torch
import torchvision.transforms.v2 as T
from numpy.typing import ArrayLike
from torch.utils.data import DataLoader
from torchvision import tv_tensors

from torch_uncertainty.datamodules import TUDataModule

from .dataset import (
    DummPixelRegressionDataset,
    DummyClassificationDataset,
    DummyRegressionDataset,
    DummySegmentationDataset,
)


class DummyClassificationDataModule(TUDataModule):
    num_channels = 1
    image_size: int = 4
    training_task = "classification"

    def __init__(
        self,
        root: str | Path,
        batch_size: int,
        num_classes: int = 2,
        num_workers: int = 1,
        eval_ood: bool = False,
        eval_shift: bool = False,
        pin_memory: bool = True,
        persistent_workers: bool = True,
        num_images: int = 2,
    ) -> None:
        super().__init__(
            root=root,
            val_split=None,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
        )

        self.eval_ood = eval_ood
        self.eval_shift = eval_shift
        self.num_classes = num_classes
        self.num_images = num_images

        self.dataset = DummyClassificationDataset
        self.ood_dataset = DummyClassificationDataset
        self.shift_dataset = DummyClassificationDataset

        self.train_transform = T.ToTensor()
        self.test_transform = T.ToTensor()

    def prepare_data(self) -> None:
        pass

    def setup(self, stage: str | None = None) -> None:
        if stage == "fit" or stage is None:
            self.train = self.dataset(
                self.root,
                num_channels=self.num_channels,
                num_classes=self.num_classes,
                image_size=self.image_size,
                transform=self.train_transform,
                num_images=self.num_images,
            )
            self.val = self.dataset(
                self.root,
                num_channels=self.num_channels,
                num_classes=self.num_classes,
                image_size=self.image_size,
                transform=self.test_transform,
                num_images=self.num_images,
            )
        elif stage == "test":
            self.test = self.dataset(
                self.root,
                num_channels=self.num_channels,
                num_classes=self.num_classes,
                image_size=self.image_size,
                transform=self.test_transform,
                num_images=self.num_images,
            )
            self.ood = self.ood_dataset(
                self.root,
                num_channels=self.num_channels,
                num_classes=self.num_classes,
                image_size=self.image_size,
                transform=self.test_transform,
                num_images=self.num_images,
            )
            self.shift = self.shift_dataset(
                self.root,
                num_channels=self.num_channels,
                num_classes=self.num_classes,
                image_size=self.image_size,
                transform=self.test_transform,
                num_images=self.num_images,
            )
            self.shift.shift_severity = 1

    def test_dataloader(self) -> DataLoader | list[DataLoader]:
        dataloader = [self._data_loader(self.test)]
        if self.eval_ood:
            dataloader.append(self._data_loader(self.ood))
        if self.eval_shift:
            dataloader.append(self._data_loader(self.shift))
        return dataloader

    def _get_train_data(self) -> ArrayLike:
        return self.train.data

    def _get_train_targets(self) -> ArrayLike:
        return np.array(self.train.targets)


class DummyRegressionDataModule(TUDataModule):
    in_features = 4
    training_task = "regression"

    def __init__(
        self,
        root: str | Path,
        batch_size: int,
        out_features: int = 2,
        num_workers: int = 1,
        pin_memory: bool = True,
        persistent_workers: bool = True,
    ) -> None:
        super().__init__(
            root=root,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
            val_split=0,
        )

        self.out_features = out_features

        self.dataset = DummyRegressionDataset
        self.ood_dataset = DummyRegressionDataset

        self.train_transform = None
        self.test_transform = None

    def prepare_data(self) -> None:
        pass

    def setup(self, stage: str | None = None) -> None:
        if stage == "fit" or stage is None:
            self.train = self.dataset(
                self.root,
                out_features=self.out_features,
                transform=self.train_transform,
            )
            self.val = self.dataset(
                self.root,
                out_features=self.out_features,
                transform=self.test_transform,
            )
        elif stage == "test":
            self.test = self.dataset(
                self.root,
                out_features=self.out_features,
                transform=self.test_transform,
            )

    def test_dataloader(self) -> DataLoader | list[DataLoader]:
        return [self._data_loader(self.test)]


class DummySegmentationDataModule(TUDataModule):
    num_channels = 3
    training_task = "segmentation"
    mean = [0.0, 0.0, 0.0]
    std = [1.0, 1.0, 1.0]

    def __init__(
        self,
        root: str | Path,
        batch_size: int,
        num_classes: int = 2,
        num_workers: int = 1,
        image_size: int = 4,
        pin_memory: bool = True,
        persistent_workers: bool = True,
        num_images: int = 2,
    ) -> None:
        super().__init__(
            root=root,
            val_split=None,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
        )

        self.num_classes = num_classes
        self.num_channels = 3
        self.num_images = num_images
        self.image_size = image_size

        self.dataset = DummySegmentationDataset

        self.train_transform = T.ToDtype(
            dtype={
                tv_tensors.Image: torch.float32,
                tv_tensors.Mask: torch.int64,
                "others": None,
            },
            scale=True,
        )
        self.test_transform = T.ToDtype(
            dtype={
                tv_tensors.Image: torch.float32,
                tv_tensors.Mask: torch.int64,
                "others": None,
            },
            scale=True,
        )

    def prepare_data(self) -> None:
        pass

    def setup(self, stage: str | None = None) -> None:
        if stage == "fit" or stage is None:
            self.train = self.dataset(
                self.root,
                num_channels=self.num_channels,
                num_classes=self.num_classes,
                image_size=self.image_size,
                transforms=self.train_transform,
                num_images=self.num_images,
            )
            self.val = self.dataset(
                self.root,
                num_channels=self.num_channels,
                num_classes=self.num_classes,
                image_size=self.image_size,
                transforms=self.test_transform,
                num_images=self.num_images,
            )
        elif stage == "test":
            self.test = self.dataset(
                self.root,
                num_channels=self.num_channels,
                num_classes=self.num_classes,
                image_size=self.image_size,
                transforms=self.test_transform,
                num_images=self.num_images,
            )

    def test_dataloader(self) -> DataLoader | list[DataLoader]:
        return [self._data_loader(self.test)]

    def _get_train_data(self) -> ArrayLike:
        return self.train.data

    def _get_train_targets(self) -> ArrayLike:
        return np.array(self.train.targets)


class DummyPixelRegressionDataModule(TUDataModule):
    num_channels = 3
    training_task = "pixel_regression"

    def __init__(
        self,
        root: str | Path,
        batch_size: int,
        output_dim: int = 2,
        num_workers: int = 1,
        image_size: int = 4,
        pin_memory: bool = True,
        persistent_workers: bool = True,
        num_images: int = 2,
    ) -> None:
        super().__init__(
            root=root,
            val_split=None,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
        )

        self.output_dim = output_dim
        self.num_channels = 3
        self.num_images = num_images
        self.image_size = image_size

        self.dataset = DummPixelRegressionDataset

        self.train_transform = T.ToDtype(
            dtype={
                tv_tensors.Image: torch.float32,
                tv_tensors.Mask: torch.float32,
                "others": None,
            },
            scale=True,
        )
        self.test_transform = T.ToDtype(
            dtype={
                tv_tensors.Image: torch.float32,
                tv_tensors.Mask: torch.float32,
                "others": None,
            },
            scale=True,
        )

    def prepare_data(self) -> None:
        pass

    def setup(self, stage: str | None = None) -> None:
        if stage == "fit" or stage is None:
            self.train = self.dataset(
                self.root,
                num_channels=self.num_channels,
                output_dim=self.output_dim,
                image_size=self.image_size,
                transforms=self.train_transform,
                num_images=self.num_images,
            )
            self.val = self.dataset(
                self.root,
                num_channels=self.num_channels,
                output_dim=self.output_dim,
                image_size=self.image_size,
                transforms=self.test_transform,
                num_images=self.num_images,
            )
        elif stage == "test":
            self.test = self.dataset(
                self.root,
                num_channels=self.num_channels,
                output_dim=self.output_dim,
                image_size=self.image_size,
                transforms=self.test_transform,
                num_images=self.num_images,
            )

    def test_dataloader(self) -> DataLoader | list[DataLoader]:
        return [self._data_loader(self.test)]

    def _get_train_data(self) -> ArrayLike:
        return self.train.data

    def _get_train_targets(self) -> ArrayLike:
        return np.array(self.train.targets)



from typing import Literal

import torch
from einops import rearrange
from torch import nn

from .std import _BasicBlock, _Bottleneck, _ResNet
from .utils import get_resnet_num_blocks

__all__ = [
    "mimo_resnet",
]


class _MIMOResNet(_ResNet):
    def __init__(
        self,
        block: type[_BasicBlock | _Bottleneck],
        num_blocks: list[int],
        in_channels: int,
        num_classes: int,
        num_estimators: int,
        conv_bias: bool,
        dropout_rate: float,
        groups: int = 1,
        style: Literal["imagenet", "cifar"] = "imagenet",
        in_planes: int = 64,
        normalization_layer: type[nn.Module] = nn.BatchNorm2d,
    ) -> None:
        super().__init__(
            block=block,
            num_blocks=num_blocks,
            in_channels=in_channels * num_estimators,
            num_classes=num_classes * num_estimators,
            conv_bias=conv_bias,
            dropout_rate=dropout_rate,
            groups=groups,
            style=style,
            in_planes=in_planes,
            normalization_layer=normalization_layer,
        )

        self.num_estimators = num_estimators

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self.training:
            x = x.repeat(self.num_estimators, 1, 1, 1)
        out = rearrange(x, "(m b) c h w -> b (m c) h w", m=self.num_estimators)
        out = super().forward(out)
        return rearrange(out, "b (m d) -> (m b) d", m=self.num_estimators)


def mimo_resnet(
    in_channels: int,
    num_classes: int,
    arch: int,
    num_estimators: int,
    conv_bias: bool = True,
    dropout_rate: float = 0.0,
    width_multiplier: float = 1.0,
    groups: int = 1,
    style: Literal["imagenet", "cifar"] = "imagenet",
    normalization_layer: type[nn.Module] = nn.BatchNorm2d,
) -> _MIMOResNet:
    block = _BasicBlock if arch in [18, 20, 34, 44, 56, 110, 1202] else _Bottleneck
    in_planes = 16 if arch in [20, 44, 56, 110, 1202] else 64
    return _MIMOResNet(
        block=block,
        num_blocks=get_resnet_num_blocks(arch),
        in_channels=in_channels,
        num_classes=num_classes,
        num_estimators=num_estimators,
        conv_bias=conv_bias,
        dropout_rate=dropout_rate,
        groups=groups,
        style=style,
        in_planes=int(in_planes * width_multiplier),
        normalization_layer=normalization_layer,
    )



from pathlib import Path

import pytest
from torchvision.datasets import ImageNet

from tests._dummies.dataset import DummyClassificationDataset
from torch_uncertainty.datamodules import ImageNetDataModule


class TestImageNetDataModule:
    """Testing the ImageNetDataModule datamodule class."""

    def test_imagenet(self):
        dm = ImageNetDataModule(root="./data/", batch_size=128, val_split=0.1)
        assert dm.dataset == ImageNet
        dm.dataset = DummyClassificationDataset
        dm.ood_dataset = DummyClassificationDataset
        dm.prepare_data()
        dm.setup()

        path = Path(__file__).parent.resolve() / "../../assets/dummy_indices.yaml"
        dm = ImageNetDataModule(root="./data/", batch_size=128, val_split=path)
        dm.dataset = DummyClassificationDataset
        dm.ood_dataset = DummyClassificationDataset
        dm.shift_dataset = DummyClassificationDataset
        dm.setup("fit")
        dm.setup("test")
        dm.train_dataloader()
        dm.val_dataloader()
        dm.test_dataloader()

        dm.val_split = None
        dm.setup("fit")
        dm.train_dataloader()
        dm.val_dataloader()
        dm.test_dataloader()

        dm.eval_ood = True
        dm.eval_shift = True
        dm.prepare_data()
        dm.setup("test")
        dm.test_dataloader()

        ImageNetDataModule(
            root="./data/",
            batch_size=128,
            val_split=path,
            rand_augment_opt="rand-m9-n1",
            basic_augment=False,
        )

        with pytest.raises(ValueError):
            dm.setup("other")

        for test_alt in ["r", "o", "a"]:
            dm = ImageNetDataModule(root="./data/", batch_size=128, test_alt=test_alt)

        with pytest.raises(ValueError):
            dm.setup()

        with pytest.raises(ValueError):
            dm = ImageNetDataModule(root="./data/", batch_size=128, test_alt="x")

        for ood_ds in ["inaturalist", "imagenet-o", "textures", "openimage-o"]:
            dm = ImageNetDataModule(root="./data/", batch_size=128, ood_ds=ood_ds)
            if ood_ds == "inaturalist":
                dm.eval_ood = True
                dm.dataset = DummyClassificationDataset
                dm.ood_dataset = DummyClassificationDataset
                dm.prepare_data()
                dm.setup("test")
                dm.test_dataloader()

        with pytest.raises(ValueError):
            dm = ImageNetDataModule(root="./data/", batch_size=128, ood_ds="other")

        for procedure in ["ViT", "A3"]:
            dm = ImageNetDataModule(
                root="./data/",
                batch_size=128,
                ood_ds="svhn",
                procedure=procedure,
            )

        with pytest.raises(ValueError):
            dm = ImageNetDataModule(root="./data/", batch_size=128, procedure="A2")

        with pytest.raises(FileNotFoundError):
            dm._verify_splits(split="test")

        with pytest.raises(FileNotFoundError):
            dm.root = Path("./tests/testlog")
            dm._verify_splits(split="test")



# ruff: noqa: F401
from .batch import MIMOBatchFormat, RepeatTarget
from .cutout import Cutout
from .image import (
    AutoContrast,
    Brightness,
    Color,
    Contrast,
    Equalize,
    Posterize,
    RandomRescale,
    Rotate,
    Sharpen,
    Shear,
    Solarize,
    Translate,
)
from .mixup import Mixup, MixupIO, RegMixup, WarpingMixup

augmentations = [
    AutoContrast,
    Equalize,
    Posterize,
    Rotate,
    Solarize,
    Shear,
    Translate,
    Contrast,
    Brightness,
    Color,
    Sharpen,
]



from collections.abc import Callable
from typing import Any

import torch
from einops import rearrange
from PIL import Image
from torchmetrics.utilities.plot import _AX_TYPE, _PLOT_OUT_TYPE
from torchvision import tv_tensors
from torchvision.datasets import Cityscapes as TVCityscapes
from torchvision.transforms.v2 import functional as F


class Cityscapes(TVCityscapes):
    def __init__(
        self,
        root: str,
        split: str = "train",
        mode: str = "fine",
        target_type: list[str] | str = "instance",
        transform: Callable[..., Any] | None = None,
        target_transform: Callable[..., Any] | None = None,
        transforms: Callable[..., Any] | None = None,
    ) -> None:
        super().__init__(
            root,
            split,
            mode,
            target_type,
            transform,
            target_transform,
            transforms,
        )
        train_id_to_color = [
            c.color for c in self.classes if (c.train_id != -1 and c.train_id != 255)
        ]
        train_id_to_color.append([0, 0, 0])
        self.train_id_to_color = torch.tensor(train_id_to_color)

    @classmethod
    def encode_target(cls, target: Image.Image) -> Image.Image:
        """Encode target image to tensor.

        Args:
            target (Image.Image): Target PIL image.

        Returns:
            torch.Tensor: Encoded target.
        """
        colored_target = F.pil_to_tensor(target)
        colored_target = rearrange(colored_target, "c h w -> h w c")
        target = torch.zeros_like(colored_target[..., :1])
        # convert target color to index
        for cityscapes_class in cls.classes:
            target[
                (colored_target == torch.tensor(cityscapes_class.id, dtype=target.dtype)).all(
                    dim=-1
                )
            ] = cityscapes_class.train_id

        return F.to_pil_image(rearrange(target, "h w c -> c h w"))

    def decode_target(self, target: torch.Tensor) -> torch.Tensor:
        """Decode target tensor to RGB tensor.

        Args:
            target (torch.Tensor): Target RGB tensor.

        Returns:
            Image.Image: Decoded target.
        """
        target[target == 255] = -1
        return self.train_id_to_color[target]

    def __getitem__(self, index: int) -> tuple[Any, Any]:
        """Get the sample at the given index.

        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is a tuple of all target types
                if ``target_type`` is a list with more
            than one item. Otherwise, target is a json object if
                ``target_type="polygon"``, else the image segmentation.
        """
        image = tv_tensors.Image(Image.open(self.images[index]).convert("RGB"))

        targets: Any = []
        for i, t in enumerate(self.target_type):
            if t == "polygon":
                target = self._load_json(self.targets[index][i])
            elif t == "semantic":
                target = tv_tensors.Mask(self.encode_target(Image.open(self.targets[index][i])))
            else:
                target = Image.open(self.targets[index][i])

            targets.append(target)

        target = tuple(targets) if len(targets) > 1 else targets[0]

        if self.transforms is not None:
            image, target = self.transforms(image, target)

        return image, target

    def plot_sample(self, index: int, ax: _AX_TYPE | None = None) -> _PLOT_OUT_TYPE:
        """Plot a sample from the dataset.

        Args:
            index: The index of the sample to plot.
            ax: Optional matplotlib axis to plot on.

        Returns:
            The axis on which the sample was plotted.
        """
        raise NotImplementedError("This method is not implemented yet.")

    @property
    def color_palette(self) -> list[tuple[int, int, int]]:
        """Return the color palette of the dataset."""
        return [c.color for c in self.classes]



from pathlib import Path

import pytest
import torch
from huggingface_hub.errors import (
    HfHubHTTPError,
    RepositoryNotFoundError,
)

from torch_uncertainty.utils import (
    csv_writer,
    distributions,
    get_version,
    hub,
    plot_hist,
)


class TestUtils:
    """Testing utils methods."""

    def test_get_version_log_success(self):
        get_version("tests/testlog", version=42)
        get_version(Path("tests/testlog"), version=42)

        get_version("tests/testlog", version=42, checkpoint=45)

    def test_getversion_log_failure(self):
        with pytest.raises(FileNotFoundError):
            get_version("tests/testlog", version=52)


class TestHub:
    """Testing hub methods."""

    def test_hub_exists(self):
        hub.load_hf("test")
        hub.load_hf("test", version=1)
        hub.load_hf("test", version=2)

    def test_hub_notexists(self):
        with pytest.raises((RepositoryNotFoundError, HfHubHTTPError)):
            hub.load_hf("tests")

        with pytest.raises((ValueError, HfHubHTTPError)):
            hub.load_hf("test", version=42)


class TestMisc:
    """Testing misc methods."""

    def test_csv_writer(self):
        root = Path(__file__).parent.resolve()
        csv_writer(root / "testlog" / "results.csv", {"a": 1.0, "b": 2.0})
        csv_writer(root / "testlog" / "results.csv", {"a": 1.0, "b": 2.0, "c": 3.0})

    def test_plot_hist(self):
        conf = [torch.rand(20), torch.rand(20)]
        plot_hist(conf, bins=10, title="test")


class TestDistributions:
    """Testing distributions methods."""

    def test_nig(self):
        dist = distributions.NormalInverseGamma(
            0.0,
            1.1,
            1.1,
            1.1,
        )
        dist = distributions.NormalInverseGamma(
            torch.tensor(0.0),
            torch.tensor(1.1),
            torch.tensor(1.1),
            torch.tensor(1.1),
        )
        _ = dist.mean, dist.mean_loc, dist.mean_variance, dist.variance_loc

    def test_get_dist_class(self):
        dist = distributions.get_dist_class("normal")
        assert dist == torch.distributions.Normal
        dist = distributions.get_dist_class("laplace")
        assert dist == torch.distributions.Laplace
        dist = distributions.get_dist_class("nig")
        assert dist == distributions.NormalInverseGamma
        dist = distributions.get_dist_class("cauchy")
        assert dist == torch.distributions.Cauchy
        dist = distributions.get_dist_class("student")
        assert dist == torch.distributions.StudentT

    def test_get_dist_estimate(self):
        dist = torch.distributions.Normal(0.0, 1.0)
        mean = distributions.get_dist_estimate(dist, "mean")
        mode = distributions.get_dist_estimate(dist, "mode")
        assert mean == mode



from typing import Literal

from torch import nn

from torch_uncertainty.models.mlp import mlp, packed_mlp
from torch_uncertainty.routines.regression import (
    RegressionRoutine,
)
from torch_uncertainty.transforms.batch import RepeatTarget

ENSEMBLE_METHODS = ["packed"]


class MLPBaseline(RegressionRoutine):
    versions = {"std": mlp, "packed": packed_mlp}

    def __init__(
        self,
        output_dim: int,
        in_features: int,
        loss: nn.Module,
        version: Literal["std", "packed"],
        hidden_dims: list[int],
        num_estimators: int | None = 1,
        dropout_rate: float = 0.0,
        alpha: float | None = None,
        gamma: int = 1,
        dist_family: str | None = None,
        dist_args: dict | None = None,
    ) -> None:
        r"""MLP baseline for regression providing support for various versions."""
        params = {
            "dropout_rate": dropout_rate,
            "in_features": in_features,
            "num_outputs": output_dim,
            "hidden_dims": hidden_dims,
            "dist_family": dist_family,
            "dist_args": dist_args,
        }

        format_batch_fn = nn.Identity()

        if version not in self.versions:
            raise ValueError(f"Unknown version: {version}")

        if version == "packed":
            params |= {
                "alpha": alpha,
                "num_estimators": num_estimators,
                "gamma": gamma,
            }
            format_batch_fn = RepeatTarget(num_repeats=num_estimators)

        model = self.versions[version](**params)

        super().__init__(
            output_dim=output_dim,
            model=model,
            loss=loss,
            dist_family=dist_family,
            is_ensemble=version in ENSEMBLE_METHODS,
            format_batch_fn=format_batch_fn,
        )
        self.save_hyperparameters(ignore=["loss"])



from collections.abc import Callable
from pathlib import Path

import pandas as pd
import torch

from .uci_classification import UCIClassificationDataset


class SpamBase(UCIClassificationDataset):
    """The SpamBase UCI classification dataset.

    Args:
        root (str): Root directory of the datasets.
        train (bool, optional): If True, creates dataset from training set,
            otherwise creates from test set.
        transform (callable, optional): A function/transform that takes in a
            numpy array and returns a transformed version.
        target_transform (callable, optional): A function/transform that takes
            in the target and transforms it.
        download (bool, optional): If true, downloads the dataset from the
            internet and puts it in root directory. If dataset is already
            downloaded, it is not downloaded again.
        binary (bool, optional): Whether to use binary classification. Defaults
            to ``True``.

    Note - License:
        The licenses of the datasets may differ from TorchUncertainty's
        license. Check before use.

    """

    md5_zip = "6159c57c5571b3c20218e32fc94e8e91"
    url = "https://archive.ics.uci.edu/static/public/94/spambase.zip"
    dataset_name = "spambase"
    filename = "spambase.data"
    num_features = 57

    def __init__(
        self,
        root: Path | str,
        transform: Callable | None = None,
        target_transform: Callable | None = None,
        binary: bool = True,
        download: bool = False,
        train: bool = True,
        test_split: float = 0.2,
        split_seed: int = 21893027,
    ) -> None:
        super().__init__(
            root,
            transform,
            target_transform,
            binary,
            download,
            train,
            test_split,
            split_seed,
        )

    def _make_dataset(self) -> None:
        """Create dataset from extracted files."""
        data = pd.read_csv(self.root / self.dataset_name / self.filename, sep=",", header=None)
        self.targets = torch.as_tensor(data[57].values, dtype=torch.long)
        self.data = torch.as_tensor(data.drop(columns=[57]).values, dtype=torch.float32)
        self.num_features = self.data.shape[1]



import logging
from pathlib import Path

import torch
from torch import nn
from torch.nn.common_types import _size_2_t
from torch.nn.modules.utils import _pair
from torchvision import tv_tensors
from torchvision.transforms import v2

from torch_uncertainty.datamodules import TUDataModule
from torch_uncertainty.datasets.segmentation import CamVid
from torch_uncertainty.transforms import RandomRescale


class CamVidDataModule(TUDataModule):
    num_channels = 3
    training_task = "segmentation"
    mean = (0.485, 0.456, 0.406)
    std = (0.229, 0.224, 0.225)

    def __init__(
        self,
        root: str | Path,
        batch_size: int,
        crop_size: _size_2_t = 640,
        eval_size: _size_2_t = (720, 960),
        group_classes: bool = True,
        basic_augment: bool = True,
        val_split: float | None = None,
        num_workers: int = 1,
        pin_memory: bool = True,
        persistent_workers: bool = True,
    ) -> None:
        r"""DataModule for the CamVid dataset.

        Args:
            root (str or Path): Root directory of the datasets.
            batch_size (int): Number of samples per batch.
            crop_size (sequence or int, optional): Desired input image and
                segmentation mask sizes during training. If :attr:`crop_size` is an
                int instead of sequence like :math:`(H, W)`, a square crop
                :math:`(\text{size},\text{size})` is made. If provided a sequence
                of length :math:`1`, it will be interpreted as
                :math:`(\text{size[0]},\text{size[1]})`. Defaults to ``640``.
            eval_size (sequence or int, optional): Desired input image and
                segmentation mask sizes during evaluation. If size is an int,
                smaller edge of the images will be matched to this number, i.e.,
                :math:`\text{height}>\text{width}`, then image will be rescaled to
                :math:`(\text{size}\times\text{height}/\text{width},\text{size})`.
                Defaults to ``(720,960)``.
            group_classes (bool, optional): Whether to group the 32 classes into
                11 superclasses. Default: ``True``.
            basic_augment (bool): Whether to apply base augmentations. Defaults to
                ``True``.
            val_split (float or None, optional): Share of training samples to use
                for validation. Defaults to ``None``.
            num_workers (int, optional): Number of dataloaders to use. Defaults to
                ``1``.
            pin_memory (bool, optional):  Whether to pin memory. Defaults to
                ``True``.
            persistent_workers (bool, optional): Whether to use persistent workers.
                Defaults to ``True``.

        Note:
            This datamodule injects the following transforms into the training and
            validation/test datasets:

            .. code-block:: python

                from torchvision.transforms import v2

                v2.Compose(
                    [
                        v2.Resize(640),
                        v2.ToDtype(
                            dtype={
                                tv_tensors.Image: torch.float32,
                                tv_tensors.Mask: torch.int64,
                                "others": None,
                            },
                            scale=True,
                        ),
                    ]
                )

            This behavior can be modified by overriding ``self.train_transform``
            and ``self.test_transform`` after initialization.
        """
        if val_split is not None:  # coverage: ignore
            logging.warning("val_split is not used for CamVidDataModule.")

        super().__init__(
            root=root,
            batch_size=batch_size,
            val_split=val_split,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
        )
        if group_classes:
            self.num_classes = 11
        else:
            self.num_classes = 32
        self.dataset = CamVid
        self.group_classes = group_classes
        self.crop_size = _pair(crop_size)
        self.eval_size = _pair(eval_size)

        if basic_augment:
            basic_transform = v2.Compose(
                [
                    RandomRescale(min_scale=0.5, max_scale=2.0, antialias=True),
                    v2.RandomCrop(
                        size=self.crop_size,
                        pad_if_needed=True,
                        fill={tv_tensors.Image: 0, tv_tensors.Mask: 255},
                    ),
                    v2.ColorJitter(brightness=0.5, contrast=0.5, saturation=0.5),
                    v2.RandomHorizontalFlip(),
                ]
            )
        else:
            basic_transform = nn.Identity()

        self.train_transform = v2.Compose(
            [
                v2.ToImage(),
                basic_transform,
                v2.ToDtype(
                    dtype={
                        tv_tensors.Image: torch.float32,
                        tv_tensors.Mask: torch.int64,
                        "others": None,
                    },
                    scale=True,
                ),
                v2.Normalize(mean=self.mean, std=self.std),
            ]
        )
        self.test_transform = v2.Compose(
            [
                v2.ToImage(),
                v2.Resize(size=self.eval_size, antialias=True),
                v2.ToDtype(
                    dtype={
                        tv_tensors.Image: torch.float32,
                        tv_tensors.Mask: torch.int64,
                        "others": None,
                    },
                    scale=True,
                ),
                v2.Normalize(mean=self.mean, std=self.std),
            ]
        )

    def prepare_data(self) -> None:  # coverage: ignore
        self.dataset(root=self.root, download=True)

    def setup(self, stage: str | None = None) -> None:
        if stage == "fit" or stage is None:
            self.train = self.dataset(
                root=self.root,
                split="train",
                group_classes=self.group_classes,
                download=False,
                transforms=self.train_transform,
            )
            self.val = self.dataset(
                root=self.root,
                split="val",
                group_classes=self.group_classes,
                download=False,
                transforms=self.test_transform,
            )
        if stage == "test" or stage is None:
            self.test = self.dataset(
                root=self.root,
                split="test",
                group_classes=self.group_classes,
                download=False,
                transforms=self.test_transform,
            )

        if stage not in ["fit", "test", None]:
            raise ValueError(f"Stage {stage} is not supported.")



import pytest
import torch

from torch_uncertainty.layers.batch_ensemble import BatchConv2d, BatchLinear


@pytest.fixture()
def feat_input() -> torch.Tensor:
    return torch.rand((4, 6))


@pytest.fixture()
def img_input() -> torch.Tensor:
    return torch.rand((5, 6, 3, 3))


class TestBatchLinear:
    """Testing the BatchLinear layer class."""

    def test_linear_one_estimator(self, feat_input: torch.Tensor):
        layer = BatchLinear(6, 2, num_estimators=1)
        print(layer)
        out = layer(feat_input)
        assert out.shape == torch.Size([4, 2])

    def test_linear_two_estimators(self, feat_input: torch.Tensor):
        layer = BatchLinear(6, 2, num_estimators=2)
        out = layer(feat_input)
        assert out.shape == torch.Size([4, 2])

    def test_linear_one_estimator_no_bias(self, feat_input: torch.Tensor):
        layer = BatchLinear(6, 2, num_estimators=1, bias=False)
        out = layer(feat_input)
        assert out.shape == torch.Size([4, 2])

    def test_convert_from_linear(self, feat_input: torch.Tensor):
        linear = torch.nn.Linear(6, 3)
        layer = BatchLinear.from_linear(linear, num_estimators=2)
        assert layer.linear.weight.shape == torch.Size([3, 6])
        assert layer.linear.bias is None
        assert layer.r_group.shape == torch.Size([2, 6])
        assert layer.s_group.shape == torch.Size([2, 3])
        assert layer.bias.shape == torch.Size([2, 3])
        out = layer(feat_input)
        assert out.shape == torch.Size([4, 3])


class TestBatchConv2d:
    """Testing the BatchConv2d layer class."""

    def test_conv_one_estimator(self, img_input: torch.Tensor):
        layer = BatchConv2d(6, 2, num_estimators=1, kernel_size=1)
        print(layer)
        out = layer(img_input)
        assert out.shape == torch.Size([5, 2, 3, 3])

    def test_conv_two_estimators(self, img_input: torch.Tensor):
        layer = BatchConv2d(6, 2, num_estimators=2, kernel_size=1)
        out = layer(img_input)
        assert out.shape == torch.Size([5, 2, 3, 3])

    def test_convert_from_conv2d(self, img_input: torch.Tensor):
        conv = torch.nn.Conv2d(6, 3, 1)
        layer = BatchConv2d.from_conv2d(conv, num_estimators=2)
        assert layer.conv.weight.shape == torch.Size([3, 6, 1, 1])
        assert layer.conv.bias is None
        assert layer.r_group.shape == torch.Size([2, 6])
        assert layer.s_group.shape == torch.Size([2, 3])
        assert layer.bias.shape == torch.Size([2, 3])
        out = layer(img_input)
        assert out.shape == torch.Size([5, 3, 3, 3])



import torch
from einops import rearrange, repeat
from torch import Tensor, nn


class RepeatTarget(nn.Module):
    def __init__(self, num_repeats: int) -> None:
        """Repeat the targets for ensemble training.

        Args:
            num_repeats: Number of times to repeat the targets.
        """
        super().__init__()

        if not isinstance(num_repeats, int):
            raise TypeError(f"num_repeats must be an integer. Got {num_repeats}.")
        if num_repeats <= 0:
            raise ValueError(f"num_repeats must be greater than 0. Got {num_repeats}.")

        self.num_repeats = num_repeats

    def forward(self, batch: tuple[Tensor, Tensor]) -> tuple[Tensor, Tensor]:
        inputs, targets = batch
        return inputs, repeat(targets, "b ... -> (m b) ...", m=self.num_repeats)


class MIMOBatchFormat(nn.Module):
    def __init__(self, num_estimators: int, rho: float = 0.0, batch_repeat: int = 1) -> None:
        """Format the batch for MIMO training.

        Args:
            num_estimators: Number of estimators.
            rho: Ratio of the correlation between the images for MIMO.
            batch_repeat: Number of times to repeat the batch.

        Reference:
            Havasi, M., et al. Training independent subnetworks for robust
            prediction. In ICLR, 2021.
        """
        super().__init__()

        if num_estimators <= 0:
            raise ValueError("num_estimators must be greater than 0.")
        if not (0.0 <= rho <= 1.0):
            raise ValueError("rho must be between 0 and 1.")
        if batch_repeat <= 0:
            raise ValueError("batch_repeat must be greater than 0.")

        self.num_estimators = num_estimators
        self.rho = rho
        self.batch_repeat = batch_repeat

    def shuffle(self, inputs: Tensor) -> Tensor:
        idx = torch.randperm(inputs.nelement(), device=inputs.device)
        return inputs.view(-1)[idx].view(inputs.size())

    def forward(self, batch: tuple[Tensor, Tensor]) -> tuple[Tensor, Tensor]:
        inputs, targets = batch
        indexes = torch.arange(0, inputs.shape[0], device=inputs.device, dtype=torch.int64).repeat(
            self.batch_repeat
        )
        main_shuffle = self.shuffle(indexes)
        threshold_shuffle = int(main_shuffle.shape[0] * (1.0 - self.rho))
        shuffle_indices = [
            torch.concat(
                [
                    self.shuffle(main_shuffle[:threshold_shuffle]),
                    main_shuffle[threshold_shuffle:],
                ],
                dim=0,
            )
            for _ in range(self.num_estimators)
        ]
        inputs = torch.stack(
            [torch.index_select(inputs, dim=0, index=indices) for indices in shuffle_indices],
            dim=0,
        )
        targets = torch.stack(
            [torch.index_select(targets, dim=0, index=indices) for indices in shuffle_indices],
            dim=0,
        )
        inputs = rearrange(inputs, "m b ... -> (m b) ...", m=self.num_estimators)
        targets = rearrange(targets, "m b -> (m b)", m=self.num_estimators)
        return inputs, targets



import pytest
import torch
from torch import nn

from torch_uncertainty.layers import BatchConv2d, BatchLinear
from torch_uncertainty.models.wrappers.batch_ensemble import BatchEnsemble, batch_ensemble


@pytest.fixture()
def img_input() -> torch.Tensor:
    return torch.rand((5, 6, 3, 3))


# Define a simple model for testing wrapper functionality (disregarding the actual BatchEnsemble architecture)
class _DummyModel(nn.Module):
    def __init__(self, in_features, out_features):
        super().__init__()
        self.conv = nn.Conv2d(in_features, out_features, 3)
        self.fc = nn.Linear(out_features, out_features)

    def forward(self, x):
        x = self.conv(x)
        x = x.flatten(1)
        return self.fc(x)


class _DummyBEModel(nn.Module):
    def __init__(self, in_features, out_features, num_estimators):
        super().__init__()
        self.conv = BatchConv2d(in_features, out_features, 3, num_estimators)
        self.fc = BatchLinear(out_features, out_features, num_estimators=num_estimators)

    def forward(self, x):
        x = self.conv(x)
        x = x.flatten(1)
        return self.fc(x)


class TestBatchEnsembleModel:
    def test_convert_layers(self):
        in_features = 6
        out_features = 4
        num_estimators = 3

        model = _DummyModel(in_features, out_features)
        wrapped_model = batch_ensemble(model, num_estimators, convert_layers=True)
        assert wrapped_model.num_estimators == num_estimators
        assert isinstance(wrapped_model.model.conv, BatchConv2d)
        assert isinstance(wrapped_model.model.fc, BatchLinear)

    def test_forward_pass(self, img_input):
        batch_size = img_input.size(0)
        in_features = img_input.size(1)
        out_features = 4
        num_estimators = 3
        model = _DummyBEModel(in_features, out_features, num_estimators)
        # with repeat_training_inputs=False
        wrapped_model = BatchEnsemble(model, num_estimators, repeat_training_inputs=False)
        # test forward pass for training
        logits = wrapped_model(img_input)
        assert logits.shape == (img_input.size(0), out_features)
        # test forward pass for evaluation
        wrapped_model.eval()
        logits = wrapped_model(img_input)
        assert logits.shape == (batch_size * num_estimators, out_features)
        # with repeat_training_inputs=True
        wrapped_model = BatchEnsemble(model, num_estimators, repeat_training_inputs=True)
        # test forward pass for training
        logits = wrapped_model(img_input)
        assert logits.shape == (batch_size * num_estimators, out_features)
        # test forward pass for evaluation
        wrapped_model.eval()
        logits = wrapped_model(img_input)
        assert logits.shape == (batch_size * num_estimators, out_features)

    def test_errors(self):
        with pytest.raises(ValueError):
            BatchEnsemble(_DummyBEModel(10, 5, 1), 0)
        with pytest.raises(ValueError):
            BatchEnsemble(_DummyModel(10, 5), 1)
        with pytest.raises(ValueError):
            BatchEnsemble(nn.Identity(), 2, convert_layers=True)



import pytest
import torch
from torch import nn

from torch_uncertainty.baselines.classification import (
    ResNetBaseline,
    WideResNetBaseline,
)


class TestMaskedBaseline:
    """Testing the MaskedResNet baseline class."""

    def test_masked_18(self):
        net = ResNetBaseline(
            num_classes=10,
            in_channels=3,
            loss=nn.CrossEntropyLoss(),
            version="masked",
            arch=18,
            style="cifar",
            num_estimators=4,
            scale=2,
            groups=1,
        )

        _ = net(torch.rand(1, 3, 32, 32))

    def test_masked_50(self):
        net = ResNetBaseline(
            num_classes=10,
            in_channels=3,
            loss=nn.CrossEntropyLoss(),
            version="masked",
            arch=50,
            style="imagenet",
            num_estimators=4,
            scale=2,
            groups=1,
        )

        _ = net(torch.rand(1, 3, 40, 40))

    def test_masked_errors(self):
        with pytest.raises(ValueError):
            _ = ResNetBaseline(
                num_classes=10,
                in_channels=3,
                loss=nn.CrossEntropyLoss(),
                version="masked",
                arch=18,
                style="cifar",
                num_estimators=4,
                scale=0.5,
                groups=1,
            )

        with pytest.raises(ValueError):
            _ = ResNetBaseline(
                num_classes=10,
                in_channels=3,
                loss=nn.CrossEntropyLoss(),
                version="masked",
                arch=18,
                style="cifar",
                num_estimators=4,
                scale=2,
                groups=0,
            )


class TestMaskedWideBaseline:
    """Testing the MaskedWideResNet baseline class."""

    def test_masked(self):
        net = WideResNetBaseline(
            num_classes=10,
            in_channels=3,
            loss=nn.CrossEntropyLoss(),
            version="masked",
            style="cifar",
            num_estimators=4,
            scale=2,
            groups=1,
        )

        _ = net(torch.rand(1, 3, 32, 32))



from typing import Literal

from torch import nn

from torch_uncertainty.models.depth.bts import bts_resnet50, bts_resnet101
from torch_uncertainty.routines import PixelRegressionRoutine


class BTSBaseline(PixelRegressionRoutine):
    single = ["std"]
    versions = {
        "std": [
            bts_resnet50,
            bts_resnet101,
        ]
    }
    archs = [50, 101]

    def __init__(
        self,
        loss: nn.Module,
        version: Literal["std"],
        arch: int,
        max_depth: float,
        dist_family: str | None = None,
        num_estimators: int = 1,
        pretrained_backbone: bool = True,
    ) -> None:
        params = {
            "dist_family": dist_family,
            "max_depth": max_depth,
            "pretrained_backbone": pretrained_backbone,
        }

        format_batch_fn = nn.Identity()

        if version not in self.versions:
            raise ValueError(f"Unknown version {version}")

        model = self.versions[version][self.archs.index(arch)](**params)
        super().__init__(
            output_dim=1,
            model=model,
            loss=loss,
            num_estimators=num_estimators,
            format_batch_fn=format_batch_fn,
            dist_family=dist_family,
        )
        self.save_hyperparameters(ignore=["loss"])



import copy

from torch import nn

from torch_uncertainty.models import EMA, SWA, deep_ensembles
from torch_uncertainty.optim_recipes import optim_cifar10_resnet18
from torch_uncertainty.post_processing import TemperatureScaler
from torch_uncertainty.routines import (
    ClassificationRoutine,
    PixelRegressionRoutine,
    RegressionRoutine,
    SegmentationRoutine,
)
from torch_uncertainty.transforms import RepeatTarget

from .model import dummy_model, dummy_segmentation_model


class DummyClassificationBaseline:
    def __new__(
        cls,
        num_classes: int,
        in_channels: int,
        loss: nn.Module,
        baseline_type: str = "single",
        optim_recipe=optim_cifar10_resnet18,
        with_feats: bool = True,
        ood_criterion: str = "msp",
        eval_ood: bool = False,
        eval_shift: bool = False,
        eval_grouping_loss: bool = False,
        calibrate: bool = False,
        save_in_csv: bool = False,
        mixtype: str = "erm",
        mixmode: str = "elem",
        dist_sim: str = "emb",
        kernel_tau_max: float = 1,
        kernel_tau_std: float = 0.5,
        mixup_alpha: float = 0,
        cutmix_alpha: float = 0,
        no_mixup_params: bool = False,
        ema: bool = False,
        swa: bool = False,
    ) -> ClassificationRoutine:
        model = dummy_model(
            in_channels=in_channels,
            num_classes=num_classes,
            with_feats=with_feats,
        )
        if ema:
            model = EMA(model, momentum=0.99)
        if swa:
            model = SWA(model, cycle_start=0, cycle_length=1)
        if not no_mixup_params:
            mixup_params = {
                "mixup_alpha": mixup_alpha,
                "cutmix_alpha": cutmix_alpha,
                "mixtype": mixtype,
                "mixmode": mixmode,
                "dist_sim": dist_sim,
                "kernel_tau_max": kernel_tau_max,
                "kernel_tau_std": kernel_tau_std,
            }
        else:
            mixup_params = None
        if baseline_type == "single":
            return ClassificationRoutine(
                num_classes=num_classes,
                model=model,
                loss=loss,
                format_batch_fn=nn.Identity(),
                log_plots=True,
                optim_recipe=optim_recipe(model),
                is_ensemble=False,
                mixup_params=mixup_params,
                ood_criterion=ood_criterion,
                eval_ood=eval_ood,
                eval_shift=eval_shift,
                eval_grouping_loss=eval_grouping_loss,
                post_processing=TemperatureScaler() if calibrate else None,
                save_in_csv=save_in_csv,
            )
        # baseline_type == "ensemble":
        model = deep_ensembles(
            [model, copy.deepcopy(model)],
            task="classification",
        )
        return ClassificationRoutine(
            num_classes=num_classes,
            model=model,
            loss=loss,
            optim_recipe=optim_recipe(model),
            format_batch_fn=RepeatTarget(2),
            log_plots=True,
            is_ensemble=True,
            ood_criterion=ood_criterion,
            eval_ood=eval_ood,
            eval_shift=eval_shift,
            eval_grouping_loss=eval_grouping_loss,
            post_processing=TemperatureScaler() if calibrate else None,
            save_in_csv=save_in_csv,
        )


class DummyRegressionBaseline:
    def __new__(
        cls,
        in_features: int,
        output_dim: int,
        loss: nn.Module,
        baseline_type: str = "single",
        optim_recipe=None,
        dist_family: str | None = "normal",
        ema: bool = False,
        swa: bool = False,
    ) -> RegressionRoutine:
        model = dummy_model(
            in_channels=in_features,
            num_classes=output_dim,
            dist_family=dist_family,
        )
        if ema:
            model = EMA(model, momentum=0.99)
        if swa:
            model = SWA(model, cycle_start=0, cycle_length=1)

        if baseline_type == "single":
            return RegressionRoutine(
                output_dim=output_dim,
                model=model,
                loss=loss,
                optim_recipe=optim_recipe(model),
                dist_family=dist_family,
            )
        # baseline_type == "ensemble":
        model = deep_ensembles(
            [model, copy.deepcopy(model)],
            task="regression",
            probabilistic=dist_family is not None,
        )
        return RegressionRoutine(
            output_dim=output_dim,
            model=model,
            loss=loss,
            is_ensemble=True,
            optim_recipe=optim_recipe(model),
            format_batch_fn=RepeatTarget(2),
            dist_family=dist_family,
        )


class DummySegmentationBaseline:
    def __new__(
        cls,
        in_channels: int,
        num_classes: int,
        image_size: int,
        loss: nn.Module,
        baseline_type: str = "single",
        optim_recipe=None,
        metric_subsampling_rate: float = 1,
        log_plots: bool = False,
        ema: bool = False,
        swa: bool = False,
    ) -> SegmentationRoutine:
        model = dummy_segmentation_model(
            in_channels=in_channels,
            num_classes=num_classes,
            image_size=image_size,
        )
        if ema:
            model = EMA(model, momentum=0.99)
        if swa:
            model = SWA(model, cycle_start=0, cycle_length=2)

        if baseline_type == "single":
            return SegmentationRoutine(
                num_classes=num_classes,
                model=model,
                loss=loss,
                format_batch_fn=None,
                optim_recipe=optim_recipe(model),
                metric_subsampling_rate=metric_subsampling_rate,
                log_plots=log_plots,
            )

        # baseline_type == "ensemble":
        model = deep_ensembles(
            [model, copy.deepcopy(model)],
            task="segmentation",
        )
        return SegmentationRoutine(
            num_classes=num_classes,
            model=model,
            loss=loss,
            format_batch_fn=RepeatTarget(2),
            optim_recipe=optim_recipe(model),
            metric_subsampling_rate=metric_subsampling_rate,
            log_plots=log_plots,
        )


class DummyPixelRegressionBaseline:
    def __new__(
        cls,
        in_channels: int,
        output_dim: int,
        image_size: int,
        loss: nn.Module,
        dist_family: str = "normal",
        baseline_type: str = "single",
        optim_recipe=None,
        ema: bool = False,
        swa: bool = False,
    ) -> PixelRegressionRoutine:
        model = dummy_segmentation_model(
            num_classes=output_dim,
            in_channels=in_channels,
            image_size=image_size,
            dist_family=dist_family,
        )
        if ema:
            model = EMA(model, momentum=0.99)
        if swa:
            model = SWA(model, cycle_start=0, cycle_length=1)

        if baseline_type == "single":
            return PixelRegressionRoutine(
                output_dim=output_dim,
                model=model,
                loss=loss,
                optim_recipe=optim_recipe(model),
                dist_family=dist_family,
            )

        # baseline_type == "ensemble":
        model = deep_ensembles(
            [model, copy.deepcopy(model)],
            task="pixel_regression",
            probabilistic=dist_family is not None,
        )
        return PixelRegressionRoutine(
            output_dim=output_dim,
            model=model,
            loss=loss,
            format_batch_fn=RepeatTarget(2),
            is_ensemble=True,
            optim_recipe=optim_recipe(model),
            dist_family=dist_family,
        )



from importlib import util
from pathlib import Path

import torch
import yaml

if util.find_spec("safetensors"):
    from safetensors.torch import load_file

    safetensors_installed = True
else:  # coverage: ignore
    safetensors_installed = False

if util.find_spec("huggingface_hub"):
    from huggingface_hub import hf_hub_download
    from huggingface_hub.errors import EntryNotFoundError

    huggingface_hub_installed = True
else:  # coverage: ignore
    huggingface_hub_installed = False


def load_hf(weight_id: str, version: int = 0) -> tuple[dict[str, torch.Tensor], dict[str, str]]:
    """Load a model from the HuggingFace hub.

    Args:
        weight_id (str): The id of the model to load.
        version (int): The id of the version when there are several on HF.

    Returns:
        tuple[dict[str, torch.Tensor], dict[str, str]]: The model weights and config.

    Note - License:
        TorchUncertainty's weights are released under the Apache 2.0 license.
    """
    if not huggingface_hub_installed:
        raise ImportError("Please install huggingface_hub to use this function.")
    if not safetensors_installed:
        raise ImportError("Please install safetensors to use this function.")
    repo_id = f"torch-uncertainty/{weight_id}"

    # Load the weights
    pickle = True
    if version is None or version == 0:
        filename = f"{weight_id}.ckpt"
    else:
        filename = f"{weight_id}_{version}.ckpt"
    try:
        weight_path = hf_hub_download(repo_id=repo_id, filename=filename)
    except EntryNotFoundError as not_pt:
        pickle = False
        filename = f"{weight_id}_{version}.safetensors"
        try:
            weight_path = hf_hub_download(repo_id=repo_id, filename=filename)
        except EntryNotFoundError:
            raise ValueError(f"Model {weight_id}_{version} not found on HuggingFace.") from not_pt

    if pickle:
        weight = torch.load(weight_path, map_location=torch.device("cpu"))
    else:
        weight = load_file(weight_path, device="cpu")

    if "state_dict" in weight:  # coverage: ignore
        weight = weight["state_dict"]

    # Load the config
    config_path = hf_hub_download(repo_id=repo_id, filename="config.yaml")
    config = yaml.safe_load(Path(config_path).read_text())
    return weight, config



import pytest
import torch
from torch import nn

from torch_uncertainty.models.lenet import batchensemble_lenet, bayesian_lenet, lenet, packed_lenet


class TestLeNet:
    """Testing the LeNet model."""

    def test_main(self):
        """Test initialization."""
        model = lenet(1, 1, norm=nn.BatchNorm2d)
        model.eval()
        model(torch.randn(1, 1, 20, 20))

        model = lenet(1, 1, norm=nn.Identity)
        model.eval()
        model(torch.randn(1, 1, 20, 20))

        batchensemble_lenet(1, 1)
        packed_lenet(1, 1)
        bayesian_lenet(1, 1)
        bayesian_lenet(
            in_channels=1,
            num_classes=1,
            num_samples=1,
            prior_sigma_1=1,
            prior_sigma_2=1,
            prior_pi=0,
            mu_init=1,
            sigma_init=1,
        )

    def test_errors(self):
        with pytest.raises(ValueError):
            lenet(1, 1, norm=nn.InstanceNorm2d)



from collections.abc import Callable
from pathlib import Path
from typing import Any

from lightning.fabric.utilities.cloud_io import get_filesystem
from lightning.pytorch import LightningDataModule, LightningModule, Trainer
from lightning.pytorch.cli import (
    ArgsType,
    LightningArgumentParser,
    LightningCLI,
    SaveConfigCallback,
)
from typing_extensions import override

from torch_uncertainty.utils.trainer import TUTrainer


class TUSaveConfigCallback(SaveConfigCallback):
    @override
    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        if self.already_saved:
            return

        if self.save_to_log_dir and stage == "fit":
            log_dir = trainer.log_dir  # this broadcasts the directory
            assert log_dir is not None
            config_path = Path(log_dir) / self.config_filename
            fs = get_filesystem(log_dir)

            if not self.overwrite:
                # check if the file exists on rank 0
                file_exists = fs.isfile(config_path) if trainer.is_global_zero else False
                # broadcast whether to fail to all ranks
                file_exists = trainer.strategy.broadcast(file_exists)
                if file_exists:  # coverage: ignore
                    raise RuntimeError(
                        f"{self.__class__.__name__} expected {config_path} to NOT exist. Aborting to avoid overwriting"
                        " results of a previous run. You can delete the previous config file,"
                        " set `LightningCLI(save_config_callback=None)` to disable config saving,"
                        ' or set `LightningCLI(save_config_kwargs={"overwrite": True})` to overwrite the config file.'
                    )

            if trainer.is_global_zero:
                fs.makedirs(log_dir, exist_ok=True)
                self.parser.save(
                    self.config,
                    config_path,
                    skip_none=False,
                    overwrite=self.overwrite,
                    multifile=self.multifile,
                )
        if trainer.is_global_zero:
            self.save_config(trainer, pl_module, stage)
            self.already_saved = True

        self.already_saved = trainer.strategy.broadcast(self.already_saved)


class TULightningCLI(LightningCLI):
    def __init__(
        self,
        model_class: (type[LightningModule] | Callable[..., LightningModule] | None) = None,
        datamodule_class: (
            type[LightningDataModule] | Callable[..., LightningDataModule] | None
        ) = None,
        save_config_callback: type[SaveConfigCallback] | None = TUSaveConfigCallback,
        save_config_kwargs: dict[str, Any] | None = None,
        trainer_class: type[Trainer] | Callable[..., Trainer] = TUTrainer,
        trainer_defaults: dict[str, Any] | None = None,
        seed_everything_default: bool | int = True,
        parser_kwargs: dict[str, Any] | dict[str, dict[str, Any]] | None = None,
        subclass_mode_model: bool = False,
        subclass_mode_data: bool = False,
        args: ArgsType = None,
        run: bool = True,
        auto_configure_optimizers: bool = True,
        eval_after_fit_default: bool = False,
    ) -> None:
        """Custom LightningCLI for torch-uncertainty.

        Args:
            model_class (type[LightningModule] | Callable[..., LightningModule] | None, optional): _description_. Defaults to None.
            datamodule_class (type[LightningDataModule] | Callable[..., LightningDataModule] | None, optional): _description_. Defaults to None.
            save_config_callback (type[SaveConfigCallback] | None, optional): _description_. Defaults to TUSaveConfigCallback.
            save_config_kwargs (dict[str, Any] | None, optional): _description_. Defaults to None.
            trainer_class (type[Trainer] | Callable[..., Trainer], optional): _description_. Defaults to Trainer.
            trainer_defaults (dict[str, Any] | None, optional): _description_. Defaults to None.
            seed_everything_default (bool | int, optional): _description_. Defaults to True.
            parser_kwargs (dict[str, Any] | dict[str, dict[str, Any]] | None, optional): _description_. Defaults to None.
            subclass_mode_model (bool, optional): _description_. Defaults to False.
            subclass_mode_data (bool, optional): _description_. Defaults to False.
            args (ArgsType, optional): _description_. Defaults to None.
            run (bool, optional): _description_. Defaults to True.
            auto_configure_optimizers (bool, optional): _description_. Defaults to True.
            eval_after_fit_default (bool, optional): _description_. Defaults to False.
        """
        self.eval_after_fit_default = eval_after_fit_default
        super().__init__(
            model_class,
            datamodule_class,
            save_config_callback,
            save_config_kwargs,
            trainer_class,
            trainer_defaults,
            seed_everything_default,
            parser_kwargs,
            subclass_mode_model,
            subclass_mode_data,
            args,
            run,
            auto_configure_optimizers,
        )

    def add_default_arguments_to_parser(self, parser: LightningArgumentParser) -> None:
        """Adds default arguments to the parser."""
        parser.add_argument(
            "--eval_after_fit",
            action="store_true",
            default=self.eval_after_fit_default,
        )
        super().add_default_arguments_to_parser(parser)

    def add_arguments_to_parser(self, parser: LightningArgumentParser) -> None:
        super().add_arguments_to_parser(parser)
        parser.link_arguments("data.eval_ood", "model.eval_ood")
        parser.link_arguments("data.eval_shift", "model.eval_shift")



from abc import ABC, abstractmethod
from importlib import util
from pathlib import Path
from typing import Literal

from lightning.pytorch.core import LightningDataModule
from numpy.typing import ArrayLike

if util.find_spec("sklearn"):
    from sklearn.model_selection import StratifiedKFold

    sklearn_installed = True
else:  # coverage: ignore
    sklearn_installed = False

from torch.utils.data import DataLoader, Dataset
from torch.utils.data.sampler import SubsetRandomSampler


class TUDataModule(ABC, LightningDataModule):
    training_task: str
    train: Dataset
    val: Dataset
    test: Dataset

    shift_severity = 1

    def __init__(
        self,
        root: str | Path,
        batch_size: int,
        val_split: float | None,
        num_workers: int,
        pin_memory: bool,
        persistent_workers: bool,
    ) -> None:
        """Abstract DataModule class.

        This class implements the basic functionality of a DataModule. It includes
        setters and getters for the datasets, dataloaders, as well as the crossval
        logic. It also provide the basic argparse arguments for the datamodules.

        Args:
            root (str): Root directory of the datasets.
            batch_size (int): Number of samples per batch.
            val_split (float): Share of samples to use for validation.
            num_workers (int): Number of workers to use for data loading.
            pin_memory (bool): Whether to pin memory.
            persistent_workers (bool): Whether to use persistent workers.
        """
        super().__init__()

        self.root = Path(root)
        self.batch_size = batch_size
        self.val_split = val_split
        self.num_workers = num_workers

        self.pin_memory = pin_memory
        self.persistent_workers = persistent_workers

    @abstractmethod
    def setup(self, stage: Literal["fit", "test"] | None = None) -> None:
        pass

    def get_train_set(self) -> Dataset:
        """Get the training set."""
        return self.train

    def get_val_set(self) -> Dataset:
        """Get the validation set."""
        return self.val

    def get_test_set(self) -> Dataset:
        """Get the test set."""
        return self.test

    def train_dataloader(self) -> DataLoader:
        r"""Get the training dataloader.

        Return:
            DataLoader: training dataloader.
        """
        return self._data_loader(self.train, shuffle=True)

    def val_dataloader(self) -> DataLoader:
        r"""Get the validation dataloader.

        Return:
            DataLoader: validation dataloader.
        """
        return self._data_loader(self.val)

    def test_dataloader(self) -> list[DataLoader]:
        r"""Get test dataloaders.

        Return:
            list[DataLoader]: test set for in distribution data
            and out-of-distribution data.
        """
        return [self._data_loader(self.test)]

    def _data_loader(self, dataset: Dataset, shuffle: bool = False) -> DataLoader:
        """Create a dataloader for a given dataset.

        Args:
            dataset (Dataset): Dataset to create a dataloader for.
            shuffle (bool, optional): Whether to shuffle the dataset. Defaults
                to False.

        Return:
            DataLoader: Dataloader for the given dataset.
        """
        return DataLoader(
            dataset,
            batch_size=self.batch_size,
            shuffle=shuffle,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            persistent_workers=self.persistent_workers,
        )

    # These two functions have to be defined in each datamodule
    # by setting the correct path to the matrix of data for each dataset.
    # It is generally "Dataset.samples" or "Dataset.data"
    # They are used for constructing cross validation splits
    def _get_train_data(self) -> ArrayLike:
        raise NotImplementedError

    def _get_train_targets(self) -> ArrayLike:
        raise NotImplementedError

    def make_cross_val_splits(self, n_splits: int = 10, train_over: int = 4) -> list:
        if not sklearn_installed:
            raise ImportError(
                "Please install torch_uncertainty with the image option"
                "to use crossval:"
                """pip install -U "torch_uncertainty[image]"."""
            )

        self.setup("fit")

        skf = StratifiedKFold(n_splits)
        cv_dm = []

        for fold, (train_idx, val_idx) in enumerate(
            skf.split(self._get_train_data(), self._get_train_targets())
        ):
            if fold >= train_over:
                break

            fold_dm = CrossValDataModule(
                root=self.root,
                train_idx=train_idx,
                val_idx=val_idx,
                datamodule=self,
                batch_size=self.batch_size,
                val_split=self.val_split,
                num_workers=self.num_workers,
                pin_memory=self.pin_memory,
                persistent_workers=self.persistent_workers,
            )
            cv_dm.append(fold_dm)

        return cv_dm


class CrossValDataModule(TUDataModule):
    def __init__(
        self,
        root: str | Path,
        train_idx: ArrayLike,
        val_idx: ArrayLike,
        datamodule: TUDataModule,
        batch_size: int,
        val_split: float,
        num_workers: int,
        pin_memory: bool,
        persistent_workers: bool,
    ) -> None:
        super().__init__(
            root=root,
            batch_size=batch_size,
            val_split=val_split,
            num_workers=num_workers,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
        )

        self.train_idx = train_idx
        self.val_idx = val_idx
        self.dm = datamodule

    def setup(self, stage: str | None = None) -> None:
        if stage == "fit" or stage is None:
            self.train = self.dm.train
            self.val = self.dm.val
        elif stage == "test":
            self.test = self.val
        else:
            raise ValueError(f"Stage {stage} not supported.")

    def _data_loader(self, dataset: Dataset, idx: ArrayLike) -> DataLoader:
        return DataLoader(
            dataset=dataset,
            sampler=SubsetRandomSampler(idx),
            shuffle=False,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            persistent_workers=self.persistent_workers,
        )

    def get_train_set(self) -> Dataset:
        """Get the training set for the current fold."""
        return self.dm.train

    def get_val_set(self) -> Dataset:
        """Get the validation set for the current fold."""
        return self.dm.val

    def get_test_set(self) -> Dataset:
        """Get the test set for the current fold."""
        return self.dm.val

    def train_dataloader(self) -> DataLoader:
        """Get the training dataloader for the current fold."""
        return self._data_loader(self.dm.get_train_set(), self.train_idx)

    def val_dataloader(self) -> DataLoader:
        """Get the validation dataloader for the current fold."""
        return self._data_loader(self.dm.get_train_set(), self.val_idx)

    def test_dataloader(self) -> DataLoader:
        """Get the test dataloader for the current fold."""
        return self._data_loader(self.dm.get_train_set(), self.val_idx)



from collections.abc import Callable
from importlib import util
from pathlib import Path
from typing import Literal

import numpy as np
from PIL import Image
from torchvision import tv_tensors
from torchvision.datasets import VisionDataset
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
    download_url,
)

if util.find_spec("cv2"):
    import cv2

    cv2_installed = True
else:  # coverage: ignore
    cv2_installed = False

if util.find_spec("h5py"):
    import h5py

    h5py_installed = True
else:  # coverage: ignore
    h5py_installed = False


class NYUv2(VisionDataset):
    root: Path
    rgb_urls = {
        "train": "http://www.doc.ic.ac.uk/~ahanda/nyu_train_rgb.tgz",
        "val": "http://www.doc.ic.ac.uk/~ahanda/nyu_test_rgb.tgz",
    }
    rgb_md5 = {
        "train": "ad124bbde47e371359caa4642a8a4611",
        "val": "f47f7c7c8a20d1210db7941c4f153b06",
    }
    depth_url = "http://horatio.cs.nyu.edu/mit/silberman/nyu_depth_v2/nyu_depth_v2_labeled.mat"
    depth_md5 = "520609c519fba3ba5ac58c8fefcc3530"

    def __init__(
        self,
        root: Path | str,
        split: Literal["train", "val"],
        transforms: Callable | None = None,
        min_depth: float = 0.0,
        max_depth: float = 10.0,
        download: bool = False,
    ):
        """NYUv2 depth dataset.

        Args:
            root (Path | str): Root directory where dataset is stored.
            split (Literal["train", "val"]): Dataset split.
            transforms (Callable | None): Transform to apply to samples & targets.
                Defaults to None.
            min_depth (float): Minimum depth value. Defaults to 1e-3.
            max_depth (float): Maximum depth value. Defaults to 10.
            download (bool): Download dataset if not found. Defaults to False.
        """
        if not cv2_installed:  # coverage: ignore
            raise ImportError(
                "The cv2 library is not installed. Please install"
                "torch_uncertainty with the image option:"
                """pip install -U "torch_uncertainty[image]"."""
            )
        if not h5py_installed:  # coverage: ignore
            raise ImportError(
                "The h5py library is not installed. Please install"
                "torch_uncertainty with the image option:"
                """pip install -U "torch_uncertainty[image]"."""
            )

        super().__init__(Path(root) / "NYUv2", transforms=transforms)
        self.min_depth = min_depth
        self.max_depth = max_depth

        if split not in ["train", "val"]:
            raise ValueError(f"split must be one of ['train', 'val']. Got {split}.")
        self.split = split

        if not self._check_integrity():
            if download:
                self._download()
            else:
                raise FileNotFoundError(
                    f"NYUv2 {split} split not found or incomplete. Set download=True to download it."
                )

        # make dataset
        path = self.root / self.split
        self.samples = sorted((path / "rgb_img").glob("**/*"))
        self.targets = sorted((path / "depth").glob("**/*"))

    def __getitem__(self, index: int):
        """Return image and target at index.

        Args:
            index (int): Index of the sample.
        """
        image = tv_tensors.Image(Image.open(self.samples[index]).convert("RGB"))
        target = Image.fromarray(
            cv2.imread(
                str(self.targets[index]),
                cv2.IMREAD_ANYCOLOR | cv2.IMREAD_ANYDEPTH,
            )
        )
        target = np.asarray(target, np.uint16)
        target = tv_tensors.Mask(target / 1e4)  # convert to meters
        target[(target <= self.min_depth) | (target > self.max_depth)] = float("nan")
        if self.transforms is not None:
            image, target = self.transforms(image, target)
        return image, target

    def __len__(self):
        """Return number of samples in dataset."""
        return len(self.samples)

    def _check_integrity(self) -> bool:
        """Check if dataset is present and complete."""
        return (
            check_integrity(
                self.root / f"nyu_{self.split}_rgb.tgz",
                self.rgb_md5[self.split],
            )
            and check_integrity(self.root / "depth.mat", self.depth_md5)
            and (self.root / self.split / "rgb_img").exists()
            and (self.root / self.split / "depth").exists()
        )

    def _download(self):
        """Download and extract dataset."""
        download_and_extract_archive(
            self.rgb_urls[self.split],
            self.root,
            extract_root=self.root / self.split / "rgb_img",
            filename=f"nyu_{self.split}_rgb.tgz",
            md5=self.rgb_md5[self.split],
        )
        if not check_integrity(self.root / "depth.mat", self.depth_md5):
            download_url(NYUv2.depth_url, self.root, "depth.mat", self.depth_md5)
        self._create_depth_files()

    def _create_depth_files(self):
        """Create depth images from the depth.mat file."""
        path = self.root / self.split
        (path / "depth").mkdir()
        samples = sorted((path / "rgb_img").glob("**/*"))
        ids = [int(p.stem.split("_")[-1]) for p in samples]
        file = h5py.File(self.root / "depth.mat", "r")
        depths = file["depths"]
        for i in range(len(depths)):
            img_id = i + 1
            if img_id in ids:
                img = (depths[i] * 1e4).astype(np.uint16).T
                Image.fromarray(img).save(path / "depth" / f"nyu_depth_{str(img_id).zfill(4)}.png")



"""
Corrupting Images with TorchUncertainty to Benchmark Robustness
===============================================================

This tutorial shows the impact of the different corruption transforms available in the
TorchUncertainty library. These corruption transforms were first proposed in the paper
Benchmarking Neural Network Robustness to Common Corruptions and Perturbations
by Dan Hendrycks and Thomas Dietterich.

For this tutorial, we will only load the corruption transforms available in 
torch_uncertainty.transforms.corruption. We also need to load utilities from
torchvision and matplotlib.
"""
# %%
from torchvision.transforms import Compose, ToTensor, Resize, CenterCrop

import matplotlib.pyplot as plt
from PIL import Image
from urllib import request

urls = [
    "https://upload.wikimedia.org/wikipedia/commons/d/d9/Carduelis_tristis_-Michigan%2C_USA_-male-8.jpg",
    "https://upload.wikimedia.org/wikipedia/commons/5/5d/Border_Collie_Blanca_y_Negra_Hembra_%28Belen%2C_Border_Collie_Los_Baganes%29.png",
    "https://upload.wikimedia.org/wikipedia/commons/f/f8/Birmakatze_Seal-Point.jpg",
    "https://upload.wikimedia.org/wikipedia/commons/a/a9/Garranos_fight.jpg",
    "https://upload.wikimedia.org/wikipedia/commons/8/8b/Cottontail_Rabbit.jpg",
]

def download_img(url, i):
    request.urlretrieve(url, f"tmp_{i}.png")
    return Image.open(f"tmp_{i}.png").convert('RGB')

images_ds = [download_img(url, i) for i, url in enumerate(urls)]


def get_images(main_corruption, index: int = 0):
    """Create an image showing the 6 levels of corruption of a given transform."""
    images = []
    for severity in range(6):
        transforms = Compose(
            [Resize(256, antialias=True), CenterCrop(256), ToTensor(), main_corruption(severity), CenterCrop(224)]
        )
        images.append(transforms(images_ds[index]).permute(1, 2, 0).numpy())
    return images


def show_images(transforms):
    """Show the effect of all given transforms."""
    num_corruptions = len(transforms)
    _, ax = plt.subplots(num_corruptions, 6, figsize=(10, int(1.5 * num_corruptions)))
    for i, transform in enumerate(transforms):
        images = get_images(transform, index=i)
        ax[i][0].text(
            -0.1,
            0.5,
            transform.__name__,
            transform=ax[i][0].transAxes,
            rotation="vertical",
            horizontalalignment="right",
            verticalalignment="center",
            fontsize=12,
        )
        for j in range(6):
            ax[i][j].imshow(images[j])
            if i == 0 and j == 0:
                ax[i][j].set_title("Original")
            elif i == 0:
                ax[i][j].set_title(f"Severity {j}")
            ax[i][j].axis("off")
    plt.show()


# %%
# 1. Noise Corruptions
# ~~~~~~~~~~~~~~~~~~~~
from torch_uncertainty.transforms.corruption import (
    GaussianNoise,
    ShotNoise,
    ImpulseNoise,
)

show_images(
    [
        GaussianNoise,
        ShotNoise,
        ImpulseNoise,
    ]
)

# %%
# 2. Blur Corruptions
# ~~~~~~~~~~~~~~~~~~~~
from torch_uncertainty.transforms.corruption import (
    MotionBlur,
    GlassBlur,
    DefocusBlur,
    ZoomBlur,
)

show_images(
    [
        GlassBlur,
        MotionBlur,
        DefocusBlur,
        ZoomBlur,
    ]
)

# %%
# 3. Weather Corruptions
# ~~~~~~~~~~~~~~~~~~~~~~
from torch_uncertainty.transforms.corruption import (
    Frost,
    Snow,
    Fog,
)

show_images(
    [
        Fog,
        Frost,
        Snow,
    ]
)

# %%
# 4. Other Corruptions
# ~~~~~~~~~~~~~~~~~~~~

from torch_uncertainty.transforms.corruption import (
    Brightness, Contrast, Elastic, JPEGCompression, Pixelate)

show_images(
    [
        Brightness,
        Contrast,
        JPEGCompression,
        Pixelate,
        Elastic,
    ]
)

# %%
# 5. Unused Corruptions
# ~~~~~~~~~~~~~~~~~~~~~

# The following corruptions are not used in the paper Benchmarking Neural Network Robustness to Common Corruptions and Perturbations.

from torch_uncertainty.transforms.corruption import (
    GaussianBlur,
    SpeckleNoise,
    Saturation,
)

show_images(
    [
        GaussianBlur,
        SpeckleNoise,
        Saturation,
    ]
)

# %%
# Reference
# ---------
#
# - **Benchmarking Neural Network Robustness to Common Corruptions and Perturbations**, Dan Hendrycks and Thomas Dietterich. `ICLR 2019 <https://arxiv.org/pdf/1903.12261>`_.



import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch.nn import init

from .sampler import CenteredGaussianMixture, TrainableDistribution


class BayesLinear(nn.Module):
    __constants__ = ["in_features", "out_features"]
    in_features: int
    out_features: int
    weight: Tensor
    lprior: Tensor
    lvposterior: Tensor

    def __init__(
        self,
        in_features: int,
        out_features: int,
        prior_sigma_1: float = 0.1,
        prior_sigma_2: float = 0.4,
        prior_pi: float = 1,
        mu_init: float = 0.0,
        sigma_init: float = -7.0,
        frozen: bool = False,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        """Bayesian Linear Layer with Mixture of Normals prior and Normal posterior.

        Args:
            in_features (int): Number of input features
            out_features (int): Number of output features
            prior_sigma_1 (float, optional): Standard deviation of the first prior
                distribution. Defaults to 0.1.
            prior_sigma_2 (float, optional): Standard deviation of the second prior
                distribution. Defaults to 0.1.
            prior_pi (float, optional): Mixture control variable. Defaults to 0.1.
            mu_init (float, optional): Initial mean of the posterior distribution.
                Defaults to 0.0.
            sigma_init (float, optional): Initial standard deviation of the
                posterior distribution. Defaults to -7.0.
            frozen (bool, optional): Whether to freeze the posterior distribution.
                Defaults to False.
            bias (bool, optional): Whether to use a bias term. Defaults to True.
            device (optional): Device to use. Defaults to None.
            dtype (optional): Data type to use. Defaults to None.

        Paper Reference:
            Blundell, Charles, et al. "Weight uncertainty in neural networks"
            ICML 2015.
        """
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.prior_sigma_1 = prior_sigma_1
        self.prior_sigma_2 = prior_sigma_2
        self.prior_pi = prior_pi
        self.mu_init = mu_init
        self.sigma_init = sigma_init
        self.frozen = frozen

        self.weight_mu = nn.Parameter(torch.empty((out_features, in_features), **factory_kwargs))
        self.weight_sigma = nn.Parameter(torch.empty((out_features, in_features), **factory_kwargs))

        if bias:
            self.bias_mu = nn.Parameter(torch.empty(out_features, **factory_kwargs))
            self.bias_sigma = nn.Parameter(torch.empty(out_features, **factory_kwargs))
        else:
            self.register_parameter("bias_mu", None)
            self.register_parameter("bias_log_sigma", None)

        self.reset_parameters()
        self.weight_sampler = TrainableDistribution(self.weight_mu, self.weight_sigma)
        if bias:
            self.bias_sampler = TrainableDistribution(self.bias_mu, self.bias_sigma)

        self.weight_prior_dist = CenteredGaussianMixture(prior_sigma_1, prior_sigma_2, prior_pi)
        if bias:
            self.bias_prior_dist = CenteredGaussianMixture(prior_sigma_1, prior_sigma_2, prior_pi)

    def reset_parameters(self) -> None:
        # TODO: change init
        init.normal_(self.weight_mu, mean=self.mu_init, std=0.1)
        init.normal_(self.weight_sigma, mean=self.sigma_init, std=0.1)

        if self.bias_mu is not None:
            init.normal_(self.bias_mu, mean=self.mu_init, std=0.1)
            init.normal_(self.bias_sigma, mean=self.sigma_init, std=0.1)

    def forward(self, inputs: Tensor) -> Tensor:
        if self.frozen:
            return self._frozen_forward(inputs)
        return self._forward(inputs)

    def _frozen_forward(self, inputs) -> Tensor:
        return F.linear(inputs, self.weight_mu, self.bias_mu)

    def _forward(self, inputs: Tensor) -> Tensor:
        weight = self.weight_sampler.sample()

        if self.bias_mu is not None:
            bias = self.bias_sampler.sample()
            bias_lposterior = self.bias_sampler.log_posterior()
            bias_lprior = self.bias_prior_dist.log_prob(bias)
        else:
            bias, bias_lposterior, bias_lprior = None, 0, 0

        self.lvposterior = self.weight_sampler.log_posterior() + bias_lposterior
        self.lprior = self.weight_prior_dist.log_prob(weight) + bias_lprior

        return F.linear(inputs, weight, bias)

    def freeze(self) -> None:
        """Freeze the layer by setting the frozen attribute to True."""
        self.frozen = True

    def unfreeze(self) -> None:
        """Unfreeze the layer by setting the frozen attribute to False."""
        self.frozen = False

    def sample(self) -> tuple[Tensor, Tensor | None]:
        """Sample the Bayesian layer's posterior."""
        weight = self.weight_sampler.sample()
        bias = self.bias_sampler.sample() if self.bias_mu is not None else None
        return weight, bias

    def extra_repr(self) -> str:
        return f"in_features={self.in_features}, out_features={self.out_features}, bias={self.bias_mu is not None}"



"""Code adapted from PixMix' paper."""

import numpy as np
from PIL import Image
from torch import nn

from torch_uncertainty.transforms import Shear, Translate, augmentations


def get_ab(beta: float) -> tuple[float, float]:
    rng = np.random.default_rng()
    if rng.uniform(low=0, high=1) < 0.5:
        a = np.float32(rng.beta(a=beta, b=1))
        b = np.float32(rng.beta(a=1, b=beta))
    else:
        a = 1 + np.float32(rng.beta(a=1, b=beta))
        b = -np.float32(rng.beta(a=1, b=beta))
    return a, b


def add(img1: Image.Image, img2: Image.Image, beta: float) -> Image.Image:
    a, b = get_ab(beta)
    img1, img2 = img1 * 2 - 1, img2 * 2 - 1
    out = a * img1 + b * img2
    return (out + 1) / 2


def multiply(img1: Image.Image, img2: Image.Image, beta: float) -> Image.Image:
    a, b = get_ab(beta)
    img1, img2 = img1 * 2, img2 * 2
    out = (img1**a) * (img2.clip(1e-37) ** b)
    return out / 2


# Summarize mixing operations
mixings = [add, multiply]


class PixMix(nn.Module):
    def __init__(
        self,
        mixing_set,
        mixing_iterations: int = 4,
        augmentation_severity: float = 3,
        mixing_severity: float = 3,
        all_ops: bool = True,
        seed: int = 12345,
    ) -> None:
        """PixMix augmentation class.

        Args:
            mixing_set (MixingSet): Dataset to be mixed with.
            mixing_iterations (int): Number of mixing iterations.
            augmentation_severity (float): Severity of augmentation.
            mixing_severity (float): Severity of mixing.
            all_ops (bool): Whether to use augmentations included in ImageNet-C.
                Defaults to True.
            seed (int): Seed for random number generator. Defaults to 12345.

        Note:
            Default arguments are set to follow original guidelines.
        """
        super().__init__()
        self.rng = np.random.default_rng(seed)
        self.mixing_set = mixing_set
        self.num_mixing_images = len(mixing_set)
        self.mixing_iterations = mixing_iterations
        self.augmentation_severity = augmentation_severity
        self.mixing_severity = mixing_severity

        if not all_ops:
            allowed_augmentations = [aug for aug in augmentations if not aug.corruption_overlap]
        else:
            allowed_augmentations = augmentations

        self.aug_instances = []
        for aug in allowed_augmentations:
            if aug in (Shear, Translate):
                self.aug_instances.append(aug(axis=0))
                self.aug_instances.append(aug(axis=1))
            else:
                self.aug_instances.append(aug())

    def __call__(self, img: Image.Image) -> np.ndarray:
        # TODO: Fix
        mixed = self.augment_input(img) if self.rng.random() < 0.5 else img

        for _ in range(self.rng.integers(low=0, high=self.mixing_iterations)):
            if self.rng.random() < 0.5:
                aug_image_copy = self._augment(img)
            else:
                aug_image_copy = self.rng.choice(self.num_mixing_images)

            # TODO: Fix
            mixed_op = self.rng.choice(mixings)
            mixed = mixed_op(np.array(mixed), np.array(aug_image_copy), self.mixing_severity)
            mixed = np.clip(mixed, 0, 1)
        return mixed

    def _augment(self, image: Image.Image) -> np.ndarray:
        op = self.rng.choice(self.aug_instances, 1)
        if op.level_type is int:
            aug_level = self._sample_int(op.pixmix_max_level)
        else:
            aug_level = self._sample_float(op.pixmix_max_level)
        return op(image.copy(), aug_level)

    def _sample_level(self) -> float:
        return self.rng.uniform(low=0.1, high=self.augmentation_severity)

    def _sample_int(self, maxval: int) -> int:
        """Helper method to scale `level` between 0 and maxval.

        Args:
            level: Level of the operation that will be between [0, maxval]
            maxval: Maximum value that the operation can have. This will be
            scaled to level/maxval.

        Returns:
            An int that results from scaling `maxval` according to `level`.
        """
        return int(self._sample_level() * maxval / 10)

    def _sample_float(self, maxval: float) -> float:
        """Helper function to scale `val` between 0 and maxval.

        Args:
            level: Level of the operation that will be in the range
                [0, `maxval`]
            maxval: Maximum value that the operation can have. This will be
                scaled to level/maxval.

        Returns:
            A float that results from scaling `maxval` according to `level`.
        """
        return float(self._sample_level()) * maxval / 10.0



from einops import rearrange
from torch import Tensor, nn


class ChannelBack(nn.Module):
    def forward(self, x: Tensor) -> Tensor:
        return rearrange(x, "b c h w -> b h w c")


class ChannelFront(nn.Module):
    def forward(self, x: Tensor) -> Tensor:
        return rearrange(x, "b h w c -> b c h w")



"""Creates the checkpoint for Hugging Face."""

import argparse
from pathlib import Path

import torch
from safetensors.torch import save_file

torch.set_grad_enabled(False)

parser = argparse.ArgumentParser(
    prog="to_hub_format",
    description="Post-process the checkpoints before the upload to HuggingFace",
)
parser.add_argument("--name", type=str, required=True, help="name of the checkpoint")
parser.add_argument("--path", type=Path, required=True, help="path to the checkpoint")
parser.add_argument("--version", type=int, default=0, help="version of the checkpoint")
parser.add_argument("--safe", action="store_true", help="whether to use safetensors")
parser.add_argument("--fp16", action="store_true", help="whether to use fp16 for the checkpoint")

args = parser.parse_args()

if not args.path.exists():
    raise ValueError("File does not exist")

dtype = torch.float16 if args.fp16 else torch.float32
model = torch.load(args.path, weights_only=True)["state_dict"]
model = {key.replace("model.", ""): val.to(device="cpu", dtype=dtype) for key, val in model.items()}

output_name = args.name
if args.version != 0:
    output_name += "_" + str(args.version)
output_name += ".ckpt" if not args.safe else ".safetensors"

if args.safe:
    save_file(model, output_name)
else:
    torch.save(model, output_name)



import pytest
import torch

from torch_uncertainty.models.wideresnet import wideresnet28x10
from torch_uncertainty.models.wideresnet.batched import (
    _BatchWideResNet,
    batched_wideresnet28x10,
)
from torch_uncertainty.models.wideresnet.masked import (
    _MaskedWideResNet,
    masked_wideresnet28x10,
)
from torch_uncertainty.models.wideresnet.mimo import (
    _MIMOWideResNet,
    mimo_wideresnet28x10,
)
from torch_uncertainty.models.wideresnet.packed import (
    _PackedWideResNet,
    packed_wideresnet28x10,
)


class TestStdWide:
    def test_main(self):
        wideresnet28x10(in_channels=1, num_classes=10, style="imagenet")

    def test_error(self):
        with pytest.raises(ValueError):
            wideresnet28x10(in_channels=1, num_classes=10, style="test")


class TestPackedResnet:
    """Testing the WideResNet packed class."""

    def test_main(self):
        packed_wideresnet28x10(
            in_channels=1,
            num_estimators=2,
            alpha=2,
            groups=1,
            gamma=1,
            num_classes=10,
            style="imagenet",
            conv_bias=False,
        )

        with pytest.raises(ValueError):
            _PackedWideResNet(
                depth=27,
                widen_factor=20,
                in_channels=3,
                num_classes=10,
                num_estimators=4,
                conv_bias=False,
                dropout_rate=0.0,
            )

        with pytest.raises(ValueError):
            packed_wideresnet28x10(
                in_channels=1,
                num_classes=10,
                num_estimators=2,
                alpha=2,
                groups=1,
                gamma=1,
                style="test",
            )


class TestMaskedWide:
    """Testing the WideResNet masked class."""

    def test_main(self):
        masked_wideresnet28x10(
            in_channels=1,
            num_classes=10,
            num_estimators=2,
            scale=2.0,
            groups=1,
            style="imagenet",
            conv_bias=False,
        )

        with pytest.raises(ValueError):
            _MaskedWideResNet(
                depth=27,
                widen_factor=20,
                in_channels=3,
                num_classes=10,
                num_estimators=4,
                conv_bias=False,
                dropout_rate=0.0,
            )

        with pytest.raises(ValueError):
            masked_wideresnet28x10(
                in_channels=1,
                num_classes=10,
                num_estimators=2,
                scale=2.0,
                groups=1,
                style="test",
            )


class TestBatchedWide:
    """Testing the WideResNet batched class."""

    def test_main(self):
        batched_wideresnet28x10(
            in_channels=1,
            num_classes=10,
            num_estimators=2,
            groups=1,
            style="imagenet",
            conv_bias=False,
        )

        with pytest.raises(ValueError):
            _BatchWideResNet(
                depth=27,
                widen_factor=20,
                in_channels=3,
                num_classes=10,
                num_estimators=4,
                conv_bias=False,
                dropout_rate=0.0,
            )

        with pytest.raises(ValueError):
            batched_wideresnet28x10(in_channels=1, num_classes=10, num_estimators=2, style="test")


class TestMIMOWide:
    """Testing the WideResNet mimo class."""

    def test_main(self):
        model = mimo_wideresnet28x10(in_channels=1, num_classes=10, num_estimators=2, style="cifar")
        model(torch.rand((2, 1, 28, 28)))

        with pytest.raises(ValueError):
            _MIMOWideResNet(
                depth=27,
                widen_factor=20,
                in_channels=3,
                num_classes=10,
                num_estimators=4,
                dropout_rate=0.0,
                conv_bias=False,
            )
        with pytest.raises(ValueError):
            mimo_wideresnet28x10(in_channels=1, num_classes=10, num_estimators=2, style="test")



# ruff: noqa: F401
from .imagenet_a import ImageNetA
from .imagenet_c import ImageNetC
from .imagenet_o import ImageNetO
from .imagenet_r import ImageNetR
from .tiny_imagenet import TinyImageNet
from .tiny_imagenet_c import TinyImageNetC



"""
Deep Evidential Regression on a Toy Example
===========================================

This tutorial provides an introduction to probabilistic regression in TorchUncertainty.

More specifically, we present Deep Evidential Regression (DER) using a practical example. We demonstrate an application of DER by tackling the toy-problem of fitting :math:`y=x^3` using a Multi-Layer Perceptron (MLP) neural network model. 
The output layer of the MLP provides a NormalInverseGamma distribution which is used to optimize the model, through its negative log-likelihood. 

DER represents an evidential approach to quantifying epistemic and aleatoric uncertainty in neural network regression models. 
This method involves introducing prior distributions over the parameters of the Gaussian likelihood function. 
Then, the MLP model estimates the parameters of this evidential distribution.

Training a MLP with DER using TorchUncertainty models and PyTorch Lightning
---------------------------------------------------------------------------

In this part, we train a neural network, based on the model and routines already implemented in TU.

1. Loading the utilities
~~~~~~~~~~~~~~~~~~~~~~~~

To train a MLP with the DER loss function using TorchUncertainty, we have to load the following modules:

- our TUTrainer 
- the model: mlp from torch_uncertainty.models.mlp
- the regression training routine from torch_uncertainty.routines
- the evidential objective: the DERLoss from torch_uncertainty.losses. This loss contains the classic NLL loss and a regularization term.
- a dataset that generates samples from a noisy cubic function: Cubic from torch_uncertainty.datasets.regression

We also need to define an optimizer using torch.optim and the neural network utils within torch.nn.
"""
# %%
import torch
from lightning import LightningDataModule
from torch import nn, optim

from torch_uncertainty import TUTrainer
from torch_uncertainty.models.mlp import mlp
from torch_uncertainty.datasets.regression.toy import Cubic
from torch_uncertainty.losses import DERLoss
from torch_uncertainty.routines import RegressionRoutine
from torch_uncertainty.layers.distributions import NormalInverseGammaLinear
from torch_uncertainty.utils.distributions import get_dist_class

# %%
# 2. The Optimization Recipe
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# We use the Adam optimizer with a rate of 5e-4.

def optim_regression(
    model: nn.Module,
    learning_rate: float = 5e-4,
):
    return optim.Adam(
        model.parameters(),
        lr=learning_rate,
        weight_decay=0,
    )


# %%
# 3. Creating the necessary variables
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# In the following, we create a trainer to train the model, the same synthetic regression 
# datasets as in the original DER paper and the model, a simple MLP with 2 hidden layers of 64 neurons each.
# Please note that this MLP finishes with a NormalInverseGammaLinear that interpret the outputs of the model
# as the parameters of a Normal Inverse Gamma distribution.

trainer = TUTrainer(accelerator="cpu", max_epochs=50) #, enable_progress_bar=False)

# dataset
train_ds = Cubic(num_samples=1000)
val_ds = Cubic(num_samples=300)

# datamodule
datamodule = LightningDataModule.from_datasets(
    train_ds, val_dataset=val_ds, test_dataset=val_ds, batch_size=32
)
datamodule.training_task = "regression"

# model
model = mlp(
    in_features=1,
    num_outputs=1,
    hidden_dims=[64, 64],
    dist_family="nig",  # Normal Inverse Gamma
)

# %%
# 4. The Loss and the Training Routine
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Next, we need to define the loss to be used during training. To do this, we
# set the weight of the regularizer of the DER Loss. After that, we define the
# training routine using the probabilistic regression training routine from
# torch_uncertainty.routines. In this routine, we provide the model, the DER
# loss, and the optimization recipe.

loss = DERLoss(reg_weight=1e-2)

routine = RegressionRoutine(
    output_dim=1,
    model=model,
    loss=loss,
    optim_recipe=optim_regression(model),
    dist_family="nig",
)

# %%
# 5. Gathering Everything and Training the Model
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Finally, we train the model using the trainer and the regression routine. We also
# test the model using the same trainer

trainer.fit(model=routine, datamodule=datamodule)
trainer.test(model=routine, datamodule=datamodule)

# %%
# 6. Testing the Model
# ~~~~~~~~~~~~~~~~~~~~
# We can now test the model by plotting the predictions and the uncertainty estimates.
# In this specific case, we can reproduce the results of the paper.

import matplotlib.pyplot as plt

with torch.no_grad():
    x = torch.linspace(-7, 7, 1000)

    dist_params = model(x.unsqueeze(-1))
    dists = get_dist_class("nig")(**dist_params)
    means = dists.loc.squeeze(1)
    variances = torch.sqrt(dists.variance_loc).squeeze(1)

fig, ax = plt.subplots(1, 1)
ax.plot(x, x**3, "--r", label="ground truth", zorder=3)
ax.plot(x, means, "-k", label="predictions")
for k in torch.linspace(0, 4, 4):
    ax.fill_between(
        x,
        means - k * variances,
        means + k * variances,
        linewidth=0,
        alpha=0.3,
        edgecolor=None,
        facecolor="blue",
        label="epistemic uncertainty" if not k else None,
    )

plt.gca().set_ylim(-150, 150)
plt.gca().set_xlim(-7, 7)
plt.legend(loc="upper left")
plt.grid()

# %%
# Reference
# ---------
#
# - **Deep Evidential Regression:** Alexander Amini, Wilko Schwarting, Ava Soleimany, & Daniela Rus. `NeurIPS 2020 <https://arxiv.org/pdf/1910.02600>`_.



from pathlib import Path

import pytest

from tests._dummies.dataset import DummyClassificationDataset
from torch_uncertainty.datamodules.abstract import (
    CrossValDataModule,
    TUDataModule,
)


class TestTUDataModule:
    """Testing the TUDataModule class."""

    def test_errors(self):
        TUDataModule.__abstractmethods__ = set()
        dm = TUDataModule("root", 128, 0.0, 4, True, True)
        with pytest.raises(NotImplementedError):
            dm.setup()
            dm._get_train_data()
            dm._get_train_targets()


class TestCrossValDataModule:
    """Testing the CrossValDataModule class."""

    def test_cv_main(self):
        TUDataModule.__abstractmethods__ = set()
        dm = TUDataModule("root", 128, 0.0, 4, True, True)
        ds = DummyClassificationDataset(Path("root"))
        dm.train = ds
        dm.val = ds
        dm.test = ds
        cv_dm = CrossValDataModule("root", [0], [1], dm, 128, 0.0, 4, True, True)

        cv_dm.setup()
        cv_dm.setup("test")

        # test abstract methods
        cv_dm.get_train_set()
        cv_dm.get_val_set()
        cv_dm.get_test_set()

        cv_dm.train_dataloader()
        cv_dm.val_dataloader()
        cv_dm.test_dataloader()

    def test_errors(self):
        TUDataModule.__abstractmethods__ = set()
        dm = TUDataModule("root", 128, 0.0, 4, True, True)
        ds = DummyClassificationDataset(Path("root"))
        dm.train = ds
        dm.val = ds
        dm.test = ds
        cv_dm = CrossValDataModule("root", [0], [1], dm, 128, 0.0, 4, True, True)
        with pytest.raises(NotImplementedError):
            cv_dm.setup()
            cv_dm._get_train_data()
            cv_dm._get_train_targets()

        with pytest.raises(ValueError):
            cv_dm.setup("other")



import copy

import torch
from torch import Tensor, nn
from torch.utils.data import DataLoader


class SWA(nn.Module):
    num_avgd_models: Tensor

    def __init__(
        self,
        model: nn.Module,
        cycle_start: int,
        cycle_length: int,
    ) -> None:
        """Stochastic Weight Averaging.

        Update the SWA model every :attr:`cycle_length` epochs starting at
        :attr:`cycle_start`. Uses the SWA model only at test time. Otherwise,
        uses the base model for training.

        Args:
            model (nn.Module): PyTorch model to be trained.
            cycle_start (int): Epoch to start SWA.
            cycle_length (int): Number of epochs between SWA updates.

        Reference:
            Izmailov, P., Podoprikhin, D., Garipov, T., Vetrov, D., & Wilson, A. G.
            (2018). Averaging Weights Leads to Wider Optima and Better Generalization.
            In UAI 2018.
        """
        super().__init__()
        _swa_checks(cycle_start, cycle_length)
        self.core_model = model
        self.cycle_start = cycle_start
        self.cycle_length = cycle_length

        self.register_buffer("num_avgd_models", torch.tensor(0, device="cpu"))
        self.swa_model = None
        self.need_bn_update = False

    @torch.no_grad()
    def update_wrapper(self, epoch: int) -> None:
        if epoch >= self.cycle_start and (epoch - self.cycle_start) % self.cycle_length == 0:
            if self.swa_model is None:
                self.swa_model = copy.deepcopy(self.core_model)
                self.num_avgd_models = torch.tensor(1)
            else:
                for swa_param, param in zip(
                    self.swa_model.parameters(),
                    self.core_model.parameters(),
                    strict=False,
                ):
                    swa_param.data += (param.data - swa_param.data) / (self.num_avgd_models + 1)
            self.num_avgd_models += 1
            self.need_bn_update = True

    def eval_forward(self, x: Tensor) -> Tensor:
        if self.swa_model is None:
            return self.core_model.forward(x)
        return self.swa_model.forward(x)

    def forward(self, x: Tensor) -> Tensor:
        if self.training:
            return self.core_model.forward(x)
        return self.eval_forward(x)

    def bn_update(self, loader: DataLoader, device) -> None:
        if self.need_bn_update and self.swa_model is not None:
            torch.optim.swa_utils.update_bn(loader, self.swa_model, device=device)
            self.need_bn_update = False


def _swa_checks(cycle_start: int, cycle_length: int) -> None:
    if cycle_start < 0:
        raise ValueError(f"`cycle_start` must be non-negative. Got {cycle_start}.")
    if cycle_length <= 0:
        raise ValueError(f"`cycle_length` must be strictly positive. Got {cycle_length}.")



from collections.abc import Callable
from pathlib import Path
from typing import Any, Literal

import torch
from torchvision.datasets import CIFAR10, CIFAR100
from torchvision.datasets.utils import (
    check_integrity,
    download_and_extract_archive,
)


class CIFAR10N(CIFAR10):
    """`CIFAR-10N <https://github.com/UCSC-REAL/cifar-10-100n>`_ Dataset.

    Args:
        root (string): Root directory of dataset where file
            ``cifar-10h-probs.npy`` exists or will be saved to if download
            is set to True.
        train (bool, optional): For API consistency, not used.
        transform (callable, optional): A function/transform that takes in
            a PIL image and returns a transformed version. E.g,
            ``transforms.RandomCrop``. Defaults to None.
        target_transform (callable, optional): A function/transform that
            takes in the target and transforms it. Defaults to None.
        download (bool, optional): If True, downloads the dataset from the
            internet and puts it in root directory. If dataset is already
            downloaded, it is not downloaded again. Defaults to False.
    """

    n_test_list = ["CIFAR-N-1.zip", "666bf3cff3a944c245f2b6f62af4b919"]
    n_url = "http://www.yliuu.com/web-cifarN/files/CIFAR-N-1.zip"
    filename = "CIFAR-N/CIFAR-10_human.pt"
    file_arg = ""

    def __init__(
        self,
        root: str | Path,
        train: bool = True,
        file_arg: Literal[
            "aggre_label",
            "worse_label",
            "random_label1",
            "random_label2",
            "random_label3",
        ] = "aggre_label",
        transform: Callable[..., Any] | None = None,
        target_transform: Callable[..., Any] | None = None,
        download: bool = False,
    ) -> None:
        super().__init__(
            Path(root),
            train=train,
            transform=transform,
            target_transform=target_transform,
            download=download,
        )

        if download:
            self.download_n()

        if not self._check_specific_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it."
            )

        self.targets = list(torch.load(self.root / self.filename)[file_arg])

    def _check_specific_integrity(self) -> bool:
        filename, md5 = self.n_test_list
        fpath = self.root / filename
        return check_integrity(fpath, md5)

    def download_n(self) -> None:
        download_and_extract_archive(
            self.n_url,
            self.root,
            filename=self.n_test_list[0],
            md5=self.n_test_list[1],
        )


class CIFAR100N(CIFAR100):
    n_test_list = ["CIFAR-N-1.zip", "666bf3cff3a944c245f2b6f62af4b919"]
    n_url = "http://www.yliuu.com/web-cifarN/files/CIFAR-N-1.zip"
    filename = "CIFAR-N/CIFAR-100_human.pt"
    file_arg = ""

    def __init__(
        self,
        root: str,
        train: bool = True,
        file_arg: Literal[
            "fine_label",
            "coarse_label",
        ] = "fine_label",
        transform: Callable[..., Any] | None = None,
        target_transform: Callable[..., Any] | None = None,
        download: bool = False,
    ) -> None:
        super().__init__(
            root,
            train=train,
            transform=transform,
            target_transform=target_transform,
            download=download,
        )
        self.root = Path(self.root)

        if download:
            self.download_n()

        if not self._check_specific_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted. You can use download=True to download it."
            )

        self.targets = list(torch.load(self.root / self.filename)[file_arg])

    def _check_specific_integrity(self) -> bool:
        filename, md5 = self.n_test_list
        fpath = self.root / filename
        return check_integrity(fpath, md5)

    def download_n(self) -> None:
        download_and_extract_archive(
            self.n_url,
            self.root,
            filename=self.n_test_list[0],
            md5=self.n_test_list[1],
        )



from pathlib import Path

import pytest
from torch import nn

from tests._dummies import (
    DummyClassificationBaseline,
    DummyClassificationDataModule,
    dummy_model,
)
from torch_uncertainty import TUTrainer
from torch_uncertainty.losses import DECLoss, ELBOLoss
from torch_uncertainty.routines import ClassificationRoutine
from torch_uncertainty.transforms import RepeatTarget


class TestClassification:
    """Testing the classification routine."""

    def test_one_estimator_binary(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=16,
            num_classes=1,
            num_images=100,
        )
        model = DummyClassificationBaseline(
            in_channels=dm.num_channels,
            num_classes=dm.num_classes,
            loss=nn.BCEWithLogitsLoss(),
            baseline_type="single",
            ood_criterion="msp",
            ema=True,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_two_estimators_binary(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=16,
            num_classes=1,
            num_images=100,
        )
        model = DummyClassificationBaseline(
            in_channels=dm.num_channels,
            num_classes=dm.num_classes,
            loss=nn.BCEWithLogitsLoss(),
            baseline_type="single",
            ood_criterion="logit",
            swa=True,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_one_estimator_two_classes(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=16,
            num_classes=2,
            num_images=100,
            eval_ood=True,
            eval_shift=True,
        )
        model = DummyClassificationBaseline(
            num_classes=dm.num_classes,
            in_channels=dm.num_channels,
            loss=nn.CrossEntropyLoss(),
            baseline_type="single",
            ood_criterion="entropy",
            eval_ood=True,
            eval_shift=True,
            no_mixup_params=True,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_one_estimator_two_classes_timm(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=16,
            num_classes=2,
            num_images=100,
            eval_ood=True,
        )
        model = DummyClassificationBaseline(
            num_classes=dm.num_classes,
            in_channels=dm.num_channels,
            loss=nn.CrossEntropyLoss(),
            baseline_type="single",
            ood_criterion="entropy",
            eval_ood=True,
            mixtype="timm",
            mixup_alpha=1.0,
            cutmix_alpha=0.5,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_one_estimator_two_classes_mixup(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=16,
            num_classes=2,
            num_images=100,
            eval_ood=True,
        )
        model = DummyClassificationBaseline(
            num_classes=dm.num_classes,
            in_channels=dm.num_channels,
            loss=nn.CrossEntropyLoss(),
            baseline_type="single",
            ood_criterion="entropy",
            eval_ood=True,
            mixtype="mixup",
            mixup_alpha=1.0,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_one_estimator_two_classes_mixup_io(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=16,
            num_classes=2,
            num_images=100,
            eval_ood=True,
        )
        model = DummyClassificationBaseline(
            num_classes=dm.num_classes,
            in_channels=dm.num_channels,
            loss=nn.CrossEntropyLoss(),
            baseline_type="single",
            ood_criterion="entropy",
            eval_ood=True,
            mixtype="mixup_io",
            mixup_alpha=1.0,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_one_estimator_two_classes_regmixup(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=16,
            num_classes=2,
            num_images=100,
            eval_ood=True,
        )
        model = DummyClassificationBaseline(
            num_classes=dm.num_classes,
            in_channels=dm.num_channels,
            loss=nn.CrossEntropyLoss(),
            baseline_type="single",
            ood_criterion="entropy",
            eval_ood=True,
            mixtype="regmixup",
            mixup_alpha=1.0,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_one_estimator_two_classes_kernel_warping_emb(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=16,
            num_classes=2,
            num_images=100,
            eval_ood=True,
        )
        model = DummyClassificationBaseline(
            num_classes=dm.num_classes,
            in_channels=dm.num_channels,
            loss=nn.CrossEntropyLoss(),
            baseline_type="single",
            ood_criterion="entropy",
            eval_ood=True,
            mixtype="kernel_warping",
            mixup_alpha=0.5,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_one_estimator_two_classes_kernel_warping_inp(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=16,
            num_classes=2,
            num_images=100,
            eval_ood=True,
        )
        model = DummyClassificationBaseline(
            num_classes=dm.num_classes,
            in_channels=dm.num_channels,
            loss=nn.CrossEntropyLoss(),
            baseline_type="single",
            ood_criterion="entropy",
            eval_ood=True,
            mixtype="kernel_warping",
            dist_sim="inp",
            mixup_alpha=0.5,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_one_estimator_two_classes_calibrated_with_ood(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True, logger=True)

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=19,  # lower than 19 it doesn't work :'(
            num_classes=2,
            num_images=100,
            eval_ood=True,
        )
        model = DummyClassificationBaseline(
            num_classes=dm.num_classes,
            in_channels=dm.num_channels,
            loss=nn.CrossEntropyLoss(),
            baseline_type="single",
            ood_criterion="energy",
            eval_ood=True,
            eval_grouping_loss=True,
            calibrate=True,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_two_estimators_two_classes_mi(self):
        trainer = TUTrainer(accelerator="cpu", fast_dev_run=True)

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=16,
            num_classes=2,
            num_images=100,
            eval_ood=True,
        )
        model = DummyClassificationBaseline(
            num_classes=dm.num_classes,
            in_channels=dm.num_channels,
            loss=DECLoss(1, 1e-2),
            baseline_type="ensemble",
            ood_criterion="mi",
            eval_ood=True,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_two_estimator_two_classes_elbo_vr_logs(self):
        trainer = TUTrainer(
            accelerator="cpu",
            max_epochs=1,
            limit_train_batches=1,
            limit_val_batches=1,
            limit_test_batches=1,
            enable_checkpointing=False,
        )

        dm = DummyClassificationDataModule(
            root=Path(),
            batch_size=16,
            num_classes=2,
            num_images=100,
            eval_ood=True,
        )
        model = DummyClassificationBaseline(
            num_classes=dm.num_classes,
            in_channels=dm.num_channels,
            loss=ELBOLoss(None, nn.CrossEntropyLoss(), kl_weight=1.0, num_samples=4),
            baseline_type="ensemble",
            ood_criterion="vr",
            eval_ood=True,
            save_in_csv=True,
        )

        trainer.fit(model, dm)
        trainer.validate(model, dm)
        trainer.test(model, dm)
        model(dm.get_test_set()[0][0])

    def test_classification_failures(self):
        # num_classes
        with pytest.raises(ValueError):
            ClassificationRoutine(num_classes=0, model=nn.Module(), loss=None)
        # single & MI
        with pytest.raises(ValueError):
            ClassificationRoutine(
                num_classes=10,
                model=nn.Module(),
                loss=None,
                is_ensemble=False,
                ood_criterion="mi",
            )
        with pytest.raises(ValueError):
            ClassificationRoutine(
                num_classes=10,
                model=nn.Module(),
                loss=None,
                ood_criterion="other",
            )

        with pytest.raises(ValueError):
            mixup_params = {"cutmix_alpha": -1}
            ClassificationRoutine(
                num_classes=10,
                model=nn.Module(),
                loss=None,
                mixup_params=mixup_params,
            )

        with pytest.raises(ValueError, match="num_calibration_bins must be at least 2, got"):
            ClassificationRoutine(
                model=nn.Identity(),
                num_classes=2,
                loss=nn.CrossEntropyLoss(),
                num_calibration_bins=0,
            )

        with pytest.raises(ValueError):
            ClassificationRoutine(
                num_classes=10,
                model=nn.Module(),
                loss=None,
                eval_grouping_loss=True,
            )

        with pytest.raises(NotImplementedError):
            ClassificationRoutine(
                num_classes=10,
                model=nn.Module(),
                loss=None,
                is_ensemble=True,
                eval_grouping_loss=True,
            )

        model = dummy_model(1, 1, 0, with_feats=False)
        with pytest.raises(ValueError):
            ClassificationRoutine(num_classes=10, model=model, loss=None, eval_grouping_loss=True)

        with pytest.raises(
            ValueError,
            match="Mixup is not supported for ensembles at training time",
        ):
            ClassificationRoutine(
                num_classes=10,
                model=nn.Module(),
                loss=None,
                mixup_params={"mixtype": "mixup"},
                format_batch_fn=RepeatTarget(2),
            )

        with pytest.raises(
            ValueError,
            match="Ensembles and post-processing methods cannot be used together. Raise an issue if needed.",
        ):
            ClassificationRoutine(
                num_classes=10,
                model=nn.Module(),
                loss=None,
                is_ensemble=True,
                post_processing=nn.Module(),
            )



import pytest

from tests._dummies.dataset import DummySegmentationDataset
from torch_uncertainty.datamodules.segmentation import CamVidDataModule
from torch_uncertainty.datasets.segmentation import CamVid


class TestCamVidDataModule:
    """Testing the CamVidDataModule datamodule."""

    def test_camvid_main(self):
        dm = CamVidDataModule(root="./data/", batch_size=128, group_classes=False)
        dm = CamVidDataModule(root="./data/", batch_size=128, basic_augment=False)

        assert dm.dataset == CamVid

        dm.dataset = DummySegmentationDataset

        dm.prepare_data()
        dm.setup()

        with pytest.raises(ValueError):
            dm.setup("xxx")

        # test abstract methods
        dm.get_train_set()
        dm.get_val_set()
        dm.get_test_set()

        dm.train_dataloader()
        dm.val_dataloader()
        dm.test_dataloader()

        dm.val_split = 0.1
        dm.prepare_data()
        dm.setup()
        dm.train_dataloader()
        dm.val_dataloader()



import matplotlib.pyplot as plt
import torch
from torch import Tensor
from torchmetrics.metric import Metric
from torchmetrics.utilities.compute import _auc_compute
from torchmetrics.utilities.data import dim_zero_cat
from torchmetrics.utilities.plot import _AX_TYPE


class AUSE(Metric):
    is_differentiable = False
    higher_is_better = False
    full_state_update = False
    plot_lower_bound = 0.0
    plot_upper_bound = 100.0
    plot_legend_name = "Sparsification Curves"

    scores: list[Tensor]
    errors: list[Tensor]

    def __init__(self, **kwargs) -> None:
        r"""The Area Under the Sparsification Error curve (AUSE) metric to evaluate
        the quality of the uncertainty estimates, i.e., how much they coincide with
        the true errors.

        Args:
            kwargs: Additional keyword arguments, see `Advanced metric settings
                <https://torchmetrics.readthedocs.io/en/stable/pages/overview.html#metric-kwargs>`_.

        Reference:
            From the paper
            `Uncertainty estimates and multi-hypotheses for optical flow <https://arxiv.org/abs/1802.07095>`_.
            In ECCV, 2018.

        Inputs:
            - :attr:`scores`: Uncertainty scores of shape :math:`(B,)`. A higher
              score means a higher uncertainty.
            - :attr:`errors`: Errors of shape :math:`(B,)`,

        where :math:`B` is the batch size.

        Note:
            A higher AUSE means a lower quality of the uncertainty estimates.
        """
        super().__init__(**kwargs)
        self.add_state("scores", default=[], dist_reduce_fx="cat")
        self.add_state("errors", default=[], dist_reduce_fx="cat")

    def update(self, scores: Tensor, errors: Tensor) -> None:
        """Store the scores and their associated errors for later computation.

        Args:
            scores (Tensor): uncertainty scores of shape :math:`(B,)`
            errors (Tensor): errors of shape :math:`(B,)`
        """
        self.scores.append(scores)
        self.errors.append(errors)

    def partial_compute(self) -> tuple[Tensor, Tensor]:
        scores = dim_zero_cat(self.scores)
        errors = dim_zero_cat(self.errors)
        if scores.shape[0] < 2:
            nan = torch.tensor([float("nan")], device=self.device)
            return nan, nan
        error_rates = _ause_rejection_rate_compute(scores, errors)
        optimal_error_rates = _ause_rejection_rate_compute(errors, errors)
        return error_rates.cpu(), optimal_error_rates.cpu()

    def compute(self) -> Tensor:
        """Compute the Area Under the Sparsification Error curve (AUSE) based
        on inputs passed to ``update``.

        Returns:
            Tensor: The AUSE.
        """
        error_rates, optimal_error_rates = self.partial_compute()
        if torch.isnan(error_rates[0]).item():
            return torch.tensor([float("nan")], device=self.device)
        num_samples = error_rates.size(0)
        x = torch.arange(0, num_samples, device=self.device) / num_samples
        y = error_rates - optimal_error_rates
        return torch.tensor([_auc_compute(x, y)])

    def plot(
        self,
        ax: _AX_TYPE | None = None,
        plot_oracle: bool = True,
        plot_value: bool = True,
    ) -> tuple[plt.Figure | None, plt.Axes]:
        """Plot the sparsification curve corresponding to the inputs passed to
        ``update``, and the oracle sparsification curve.

        Args:
            ax (Axes | None, optional): An matplotlib axis object. If provided
                will add plot to this axis. Defaults to None.
            plot_oracle (bool, optional): Whether to plot the oracle
                sparsification curve. Defaults to True.
            plot_value (bool, optional): Whether to plot the AUSE value.
                Defaults to True.

        Returns:
            tuple[[Figure | None], Axes]: Figure object and Axes object
        """
        fig, ax = plt.subplots() if ax is None else (None, ax)

        # Computation of AUSEC
        error_rates, optimal_error_rates = self.partial_compute()
        num_samples = error_rates.size(0)
        x = torch.arange(num_samples) / num_samples
        y = error_rates - optimal_error_rates

        ausec = _auc_compute(x, y).cpu().item()

        rejection_rates = torch.arange(num_samples) / num_samples * 100

        ax.plot(
            rejection_rates,
            error_rates * 100,
            label="Model",
        )
        if plot_oracle:
            ax.plot(
                rejection_rates,
                optimal_error_rates * 100,
                label="Oracle",
            )

        ax.set_xlabel("Rejection Rate (%)")
        ax.set_ylabel("Error Rate (%)")
        ax.set_xlim(self.plot_lower_bound, self.plot_upper_bound)
        ax.set_ylim(self.plot_lower_bound, self.plot_upper_bound)
        ax.legend(loc="upper right")

        if plot_value:
            ax.text(
                0.02,
                0.02,
                f"AUSEC={ausec:.03}",
                color="black",
                ha="left",
                va="bottom",
                transform=ax.transAxes,
            )

        return fig, ax


def _ause_rejection_rate_compute(
    scores: Tensor,
    errors: Tensor,
) -> Tensor:
    """Compute the cumulative error rates for a given set of scores and errors.

    Args:
        scores (Tensor): uncertainty scores of shape :math:`(B,)`
        errors (Tensor): errors of shape :math:`(B,)`
    """
    num_samples = errors.size(0)

    order = scores.argsort()
    errors = errors[order]

    error_rates = torch.zeros(num_samples + 1)
    error_rates[0] = errors.sum()
    error_rates[1:] = errors.cumsum(dim=-1).flip(0)
    return error_rates / error_rates[0]



import matplotlib.pyplot as plt
import pytest
import torch

from torch_uncertainty.metrics.classification import (
    AURC,
    CovAtxRisk,
    RiskAtxCov,
)


class TestAURC:
    """Testing the AURC metric class."""

    def test_compute_binary(self) -> None:
        probs = torch.Tensor([0.1, 0.2, 0.3, 0.4, 0.2])
        targets = torch.Tensor([1, 1, 1, 1, 1])
        metric = AURC()
        assert metric(probs, targets).item() == pytest.approx(1)
        targets = torch.Tensor([0, 0, 0, 0, 0])
        metric = AURC()
        assert metric(probs, targets).item() == pytest.approx(0)
        targets = torch.Tensor([0, 0, 1, 1, 0])
        metric = AURC()
        value = (0 * 0.4 + 0.25 * 0.2 / 2 + 0.25 * 0.2 + 0.15 * 0.2 / 2) / 0.8
        assert metric(probs, targets).item() == pytest.approx(value)

        metric = AURC()
        assert torch.isnan(metric(torch.Tensor([0.0]), torch.Tensor([1])))

    def test_compute_multiclass(self) -> None:
        probs = torch.Tensor([[0.1, 0.9], [0.2, 0.8], [0.3, 0.7], [0.4, 0.6], [0.2, 0.8]])
        targets = torch.Tensor([1, 1, 1, 1, 1]).long()
        metric = AURC()
        assert metric(probs, targets).item() == pytest.approx(0)
        targets = torch.Tensor([0, 0, 0, 0, 0])
        metric = AURC()
        assert metric(probs, targets).item() == pytest.approx(1)
        targets = torch.Tensor([1, 1, 0, 0, 1])
        metric = AURC()
        value = (0 * 0.4 + 0.25 * 0.2 / 2 + 0.25 * 0.2 + 0.15 * 0.2 / 2) / 0.8
        assert metric(probs, targets).item() == pytest.approx(value)

    def test_compute_nan(self) -> None:
        probs = torch.Tensor([[0.1, 0.9]])
        targets = torch.Tensor([1]).long()
        metric = AURC()
        assert torch.isnan(metric(probs, targets)).all()

    def test_plot(self) -> None:
        scores = torch.Tensor([0.2, 0.1, 0.5, 0.3, 0.4])
        values = torch.Tensor([0.1, 0.2, 0.3, 0.4, 0.5])
        metric = AURC()
        metric.update(scores, values)
        fig, ax = metric.plot()
        assert isinstance(fig, plt.Figure)
        assert isinstance(ax, plt.Axes)
        assert ax.get_xlabel() == "Coverage (%)"
        assert ax.get_ylabel() == "Risk - Error Rate (%)"
        plt.close(fig)

        metric = AURC()
        assert metric(torch.zeros(1), torch.zeros(1)).isnan()

        metric = AURC()
        metric.update(scores, values)
        fig, ax = metric.plot(plot_value=False)
        assert isinstance(fig, plt.Figure)
        assert isinstance(ax, plt.Axes)
        assert ax.get_xlabel() == "Coverage (%)"
        assert ax.get_ylabel() == "Risk - Error Rate (%)"
        plt.close(fig)


class TestCovAtxRisk:
    """Testing the CovAtxRisk metric class."""

    def test_compute_zero(self) -> None:
        probs = torch.Tensor([[0.9, 0.1], [0.8, 0.2], [0.7, 0.3], [0.6, 0.4], [0.8, 0.2]])
        targets = torch.Tensor([1, 1, 1, 1, 1])
        metric = CovAtxRisk(risk_threshold=0.5)
        # no cov for given risk
        assert torch.isnan(metric(probs, targets))

        probs = torch.Tensor([0.1, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.48, 0.49])
        targets = torch.Tensor([1, 0, 1, 1, 1, 0, 0, 0, 1])
        metric = CovAtxRisk(risk_threshold=0.55)
        # multiple cov for given risk
        assert metric(probs, targets) == pytest.approx(8 / 9)

        probs = torch.Tensor([0.1, 0.2, 0.3, 0.4, 0.2])
        targets = torch.Tensor([0, 0, 1, 1, 1])
        metric = CovAtxRisk(risk_threshold=0.5)
        assert metric(probs, targets) == pytest.approx(4 / 5)

        targets = torch.Tensor([0, 0, 1, 1, 0])
        metric = CovAtxRisk(risk_threshold=0.5)
        assert metric(probs, targets) == 1

        metric = CovAtxRisk(risk_threshold=0.5)
        assert metric(torch.zeros(0), torch.zeros(0)).isnan()

    def test_errors(self):
        with pytest.raises(TypeError, match="Expected threshold to be of type float"):
            CovAtxRisk(risk_threshold="0.5")
        with pytest.raises(ValueError, match="Threshold should be in the range"):
            CovAtxRisk(risk_threshold=-0.5)


class TestRiskAtxCov:
    """Testing the RiskAtxCov metric class."""

    def test_compute_zero(self) -> None:
        probs = torch.Tensor([[0.9, 0.1], [0.8, 0.2], [0.7, 0.3], [0.6, 0.4], [0.8, 0.2]])
        targets = torch.Tensor([1, 1, 1, 1, 1])
        metric = RiskAtxCov(cov_threshold=0.5)
        assert metric(probs, targets) == 1

        probs = torch.Tensor([0.1, 0.2, 0.3, 0.4, 0.2])
        targets = torch.Tensor([0, 0, 1, 1, 1])
        metric = RiskAtxCov(cov_threshold=0.5)
        assert metric(probs, targets) == pytest.approx(1 / 3)

        probs = torch.Tensor([0.1, 0.19, 0.3, 0.15, 0.4, 0.2])
        targets = torch.Tensor([0, 0, 1, 0, 1, 1])
        metric = RiskAtxCov(cov_threshold=0.5)
        assert metric(probs, targets) == 0

        probs = torch.Tensor([0.1, 0.2, 0.3, 0.15, 0.4, 0.2])
        targets = torch.Tensor([0, 0, 1, 0, 1, 1])
        metric = RiskAtxCov(cov_threshold=0.55)
        assert metric(probs, targets) == 1 / 4



from collections import OrderedDict

from lightning.pytorch.loops.evaluation_loop import _EvaluationLoop
from lightning.pytorch.trainer.connectors.logger_connector.result import (
    _OUT_DICT,
)
from rich import get_console
from rich.console import Group
from rich.table import Table
from torch import Tensor

PERCENTAGE_METRICS = [
    "Acc",
    "AUPR",
    "AUROC",
    "FPR95",
    "Cov@5Risk",
    "Risk@80Cov",
    "pixAcc",
    "mIoU",
    "AURC",
    "AUGRC",
    "mAcc",
]


def _add_row(table: Table, metric_name: str, value: Tensor) -> None:
    if metric_name in PERCENTAGE_METRICS:
        value = value * 100
        table.add_row(metric_name, f"{value.item():.3f}%")
    else:
        table.add_row(metric_name, f"{value.item():.5f}")


class TUEvaluationLoop(_EvaluationLoop):
    @staticmethod
    def _print_results(results: list[_OUT_DICT], stage: str) -> None:
        # test/cls: Classification Metrics
        # test/cal: Calibration Metrics
        # ood: OOD Detection Metrics
        # shift: Distribution shift Metrics
        # test/sc: Selective Classification Metrics
        # test/post: Post-Processing Metrics
        # test/seg: Segmentation Metrics

        metrics = {}
        for result in results:
            for key, value in result.items():
                if key.startswith("test/cls"):
                    if "cls" not in metrics:
                        metrics["cls"] = {}
                    metric_name = key.split("/")[-1]
                    metrics["cls"].update({metric_name: value})
                elif key.startswith("test/cal"):
                    if "cal" not in metrics:
                        metrics["cal"] = {}
                    metric_name = key.split("/")[-1]
                    metrics["cal"].update({metric_name: value})
                elif key.startswith("ood"):
                    if "ood" not in metrics:
                        metrics["ood"] = {}
                    metric_name = key.split("/")[-1]
                    metrics["ood"].update({metric_name: value})
                elif key.startswith("shift"):
                    if "shift" not in metrics:
                        metrics["shift"] = {}
                    metric_name = key.split("/")[-1]
                    metrics["shift"].update({metric_name: value})
                elif key.startswith("test/sc"):
                    if "sc" not in metrics:
                        metrics["sc"] = {}
                    metric_name = key.split("/")[-1]
                    metrics["sc"].update({metric_name: value})
                elif key.startswith("test/post"):
                    if "post" not in metrics:
                        metrics["post"] = {}
                    metric_name = key.split("/")[-1]
                    metrics["post"].update({metric_name: value})
                elif key.startswith("test/seg"):
                    if "seg" not in metrics:
                        metrics["seg"] = {}
                    metric_name = key.split("/")[-1]
                    metrics["seg"].update({metric_name: value})
                elif key.startswith("test/reg"):
                    if "reg" not in metrics:
                        metrics["reg"] = {}
                    metric_name = key.split("/")[-1]
                    metrics["reg"].update({metric_name: value})

        tables = []

        first_col_name = f"{stage.capitalize()} metric"

        if "cls" in metrics:
            table = Table()
            table.add_column(first_col_name, justify="center", style="cyan", width=12)
            table.add_column("Classification", justify="center", style="magenta", width=25)
            cls_metrics = OrderedDict(sorted(metrics["cls"].items()))
            for metric_name, value in cls_metrics.items():
                _add_row(table, metric_name, value)
            tables.append(table)

        if "seg" in metrics:
            table = Table()
            table.add_column(first_col_name, justify="center", style="cyan", width=12)
            table.add_column("Segmentation", justify="center", style="magenta", width=25)
            seg_metrics = OrderedDict(sorted(metrics["seg"].items()))
            for metric_name, value in seg_metrics.items():
                _add_row(table, metric_name, value)
            tables.append(table)

        if "reg" in metrics:
            table = Table()
            table.add_column(first_col_name, justify="center", style="cyan", width=12)
            table.add_column("Regression", justify="center", style="magenta", width=25)
            reg_metrics = OrderedDict(sorted(metrics["reg"].items()))
            for metric_name, value in reg_metrics.items():
                _add_row(table, metric_name, value)
            tables.append(table)

        if "cal" in metrics:
            table = Table()
            table.add_column(first_col_name, justify="center", style="cyan", width=12)
            table.add_column("Calibration", justify="center", style="magenta", width=25)
            cal_metrics = OrderedDict(sorted(metrics["cal"].items()))
            for metric_name, value in cal_metrics.items():
                _add_row(table, metric_name, value)
            tables.append(table)

        if "ood" in metrics:
            table = Table()
            table.add_column(first_col_name, justify="center", style="cyan", width=12)
            table.add_column("OOD Detection", justify="center", style="magenta", width=25)
            ood_metrics = OrderedDict(sorted(metrics["ood"].items()))
            for metric_name, value in ood_metrics.items():
                _add_row(table, metric_name, value)
            tables.append(table)

        if "sc" in metrics:
            table = Table()
            table.add_column(first_col_name, justify="center", style="cyan", width=12)
            table.add_column(
                "Selective Classification",
                justify="center",
                style="magenta",
                width=25,
            )
            sc_metrics = OrderedDict(sorted(metrics["sc"].items()))
            for metric_name, value in sc_metrics.items():
                _add_row(table, metric_name, value)
            tables.append(table)

        if "post" in metrics:
            table = Table()
            table.add_column(first_col_name, justify="center", style="cyan", width=12)
            table.add_column("Post-Processing", justify="center", style="magenta", width=25)
            post_metrics = OrderedDict(sorted(metrics["post"].items()))
            for metric_name, value in post_metrics.items():
                _add_row(table, metric_name, value)
            tables.append(table)

        if "shift" in metrics:
            table = Table()
            table.add_column(first_col_name, justify="center", style="cyan", width=12)
            shift_severity = int(metrics["shift"]["shift_severity"])
            table.add_column(
                f"Distribution Shift lvl{shift_severity}",
                justify="center",
                style="magenta",
                width=25,
            )
            shift_metrics = OrderedDict(sorted(metrics["shift"].items()))
            for metric_name, value in shift_metrics.items():
                if metric_name == "shift_severity":
                    continue
                _add_row(table, metric_name, value)
            tables.append(table)

        console = get_console()
        group = Group(*tables)
        console.print(group)



import pytest
import torch

from torch_uncertainty.layers.channel_layer_norm import ChannelLayerNorm
from torch_uncertainty.layers.filter_response_norm import (
    FilterResponseNorm1d,
    FilterResponseNorm2d,
    FilterResponseNorm3d,
    _FilterResponseNormNd,
)
from torch_uncertainty.layers.mc_batch_norm import (
    MCBatchNorm1d,
    MCBatchNorm2d,
    MCBatchNorm3d,
)


class TestFilterResponseNorm:
    """Testing the FRN2d layer."""

    def test_main(self):
        """Test initialization."""
        frn = FilterResponseNorm2d(1)
        frn(torch.randn(1, 1, 20, 20))
        FilterResponseNorm1d(1)
        FilterResponseNorm3d(1)

    def test_errors(self):
        """Test errors."""
        with pytest.raises(ValueError):
            _FilterResponseNormNd(-1, 1)
        with pytest.raises(ValueError):
            FilterResponseNorm2d(0)
        with pytest.raises(ValueError):
            FilterResponseNorm2d(1.5)


class TestMCBatchNorm:
    """Testing the MCBatchNorm layers."""

    def test_main(self):
        """Test initialization."""
        bn = MCBatchNorm1d(1, 1)
        bn(torch.randn(1, 1, 2))
        bn = MCBatchNorm2d(1, 1)
        bn(torch.randn(1, 1, 1, 2))
        bn = MCBatchNorm3d(1, 1)
        bn(torch.randn(1, 1, 1, 1, 2))

    def test_errors(self):
        """Test errors."""
        with pytest.raises(ValueError):
            MCBatchNorm2d(1, 0)
        with pytest.raises(ValueError):
            MCBatchNorm2d(1, 1.5)

        layer1d = MCBatchNorm1d(1, 1)
        layer2d = MCBatchNorm2d(1, 1)
        layer3d = MCBatchNorm3d(1, 1)
        with pytest.raises(ValueError):
            layer1d(torch.randn(1, 1, 1, 20))
        with pytest.raises(ValueError):
            layer2d(torch.randn(1, 1, 1, 1, 20))
        with pytest.raises(ValueError):
            layer3d(torch.randn(1, 1, 1, 1, 1, 20))


class TestChannelLayerNorm:
    """Testing the FRN2d layer."""

    def test_main(self):
        """Test initialization."""
        cln = ChannelLayerNorm(1)
        cln(torch.randn(1, 1, 4, 4))
        cln = ChannelLayerNorm(18)
        cln(torch.randn(1, 18, 2, 3))



from copy import deepcopy
from typing import Literal

import torch
from torch import Tensor, nn
from torch.utils.data import DataLoader, Dataset

from torch_uncertainty.layers.mc_batch_norm import MCBatchNorm2d
from torch_uncertainty.post_processing import PostProcessing


class MCBatchNorm(PostProcessing):
    counter: int = 0
    mc_batch_norm_layers: list[MCBatchNorm2d] = []
    trained = False

    def __init__(
        self,
        model: nn.Module | None = None,
        num_estimators: int = 16,
        convert: bool = True,
        mc_batch_size: int = 32,
        device: Literal["cpu", "cuda"] | torch.device | None = None,
    ) -> None:
        """Monte Carlo Batch Normalization wrapper.

        Args:
            model (nn.Module): model to be converted.
            num_estimators (int): number of estimators.
            convert (bool): whether to convert the model.
            mc_batch_size (int, optional): Monte Carlo batch size. Defaults to 32.
            device (Literal["cpu", "cuda"] | torch.device | None, optional): device.
                Defaults to None.

        Note:
            This wrapper will be stochastic in eval mode only.

        Reference:
            Teye M, Azizpour H, Smith K. Bayesian uncertainty estimation for
            batch normalized deep networks. In ICML 2018.
        """
        super().__init__()
        self.mc_batch_size = mc_batch_size
        self.convert = convert
        self.num_estimators = num_estimators
        self.device = device

        if model is not None:
            self._setup_model(model)

    def _setup_model(self, model):
        _mcbn_checks(model, self.num_estimators, self.mc_batch_size, self.convert)
        self.model = deepcopy(model)  # Is it necessary?
        self.model = self.model.eval()
        if self.convert:
            self._convert()
            if not has_mcbn(self.model):
                raise ValueError("model does not contain any MCBatchNorm2d after conversion.")

    def set_model(self, model: nn.Module) -> None:
        self.model = model
        self._setup_model(model)

    def fit(self, dataset: Dataset) -> None:
        """Fit the model on the dataset.

        Args:
            dataset (Dataset): dataset to be used for fitting.

        Note:
            This method is used to populate the MC BatchNorm layers.
            Use the training dataset.
        """
        self.dl = DataLoader(dataset, batch_size=self.mc_batch_size, shuffle=True)
        self.counter = 0
        self.reset_counters()
        self.set_accumulate(True)
        self.eval()
        for x, _ in self.dl:
            self.model(x.to(self.device))
            self.raise_counters()
            if self.counter == self.num_estimators:
                self.set_accumulate(False)
                self.trained = True
                return
        raise ValueError("The dataset is too small to populate the MC BatchNorm statistics.")

    def _est_forward(self, x: Tensor) -> Tensor:
        """Forward pass of a single estimator."""
        logit = self.model(x)
        self.raise_counters()
        return logit

    def forward(
        self,
        x: Tensor,
    ) -> Tensor:
        if self.training:
            return self.model(x)
        if not self.trained:
            raise RuntimeError("MCBatchNorm has not been trained. Call .fit() first.")
        self.reset_counters()
        return torch.cat([self._est_forward(x) for _ in range(self.num_estimators)], dim=0)

    def _convert(self) -> None:
        """Convert all BatchNorm2d layers to MCBatchNorm2d layers."""
        self.replace_layers(self.model)

    def reset_counters(self) -> None:
        """Reset all counters to 0."""
        self.counter = 0
        for layer in self.mc_batch_norm_layers:
            layer.set_counter(0)

    def raise_counters(self) -> None:
        """Raise all counters by 1."""
        self.counter += 1
        for layer in self.mc_batch_norm_layers:
            layer.set_counter(self.counter)

    def set_accumulate(self, accumulate: bool) -> None:
        """Set the accumulate flag for all MCBatchNorm2d layers.

        Args:
            accumulate (bool): accumulate flag.
        """
        for layer in self.mc_batch_norm_layers:
            layer.accumulate = accumulate

    def replace_layers(self, model: nn.Module) -> None:
        """Replace all BatchNorm2d layers with MCBatchNorm2d layers.

        Args:
            model (nn.Module): model to be converted.
        """
        for name, module in model.named_children():
            if len(list(module.children())) > 0:
                self.replace_layers(module)

            if isinstance(module, nn.BatchNorm2d):
                mc_layer = MCBatchNorm2d(
                    num_features=module.num_features,
                    num_estimators=self.num_estimators,
                    eps=module.eps,
                    momentum=module.momentum,
                    affine=module.affine,
                    track_running_stats=module.track_running_stats,
                    device=module.weight.device,
                    dtype=module.weight.dtype,
                )
                mc_layer.training = module.training
                mc_layer.weight = module.weight
                mc_layer.bias = module.bias
                setattr(model, name, mc_layer)

                # Save pointers to the MC BatchNorm layers
                self.mc_batch_norm_layers.append(mc_layer)


def has_mcbn(model: nn.Module) -> bool:
    """Check if the model contains any MCBatchNorm2d layers."""
    return any(isinstance(module, MCBatchNorm2d) for module in model.modules())


def _mcbn_checks(model, num_estimators, mc_batch_size, convert):
    if num_estimators < 1 or not isinstance(num_estimators, int):
        raise ValueError(f"num_estimators must be a positive integer, got {num_estimators}.")
    if mc_batch_size < 1 or not isinstance(mc_batch_size, int):
        raise ValueError(f"mc_batch_size must be a positive integer, got {mc_batch_size}.")
    if not convert and not has_mcbn(model):
        raise ValueError("model does not contain any MCBatchNorm2d nor is not to be converted.")



from .base import ImageNetVariation


class ImageNetA(ImageNetVariation):
    url = "https://people.eecs.berkeley.edu/~hendrycks/imagenet-a.tar"
    filename = "imagenet-a.tar"
    tgz_md5 = "c3e55429088dc681f30d81f4726b6595"
    dataset_name = "imagenet-a"



import torch
from torch.utils.data import Dataset


class AggregatedDataset(Dataset):
    """A class to help aggregating datasets.

    Virtually interlace multiple copies of a dataset to train ensembles with
    different batch orders.

    Args:
        dataset: The dataset to be interlaced.
        n_dataloaders: The number of dataloaders to be used for training.
    """

    def __init__(self, dataset: Dataset, n_dataloaders: int) -> None:
        super().__init__()
        self.dataset = dataset
        self.n_dataloaders = n_dataloaders
        self.dataset_size = len(dataset)
        self.offset = self.dataset_size // self.n_dataloaders

    def __getitem__(self, idx: int) -> tuple[torch.Tensor, torch.Tensor]:
        """Get the samples and targets of the dataset.

        Args:
            idx: The index of the sample.
        """
        inputs, targets = zip(
            *[
                self.dataset[(idx + i * self.offset) % self.dataset_size]
                for i in range(self.n_dataloaders)
            ],
            strict=True,
        )
        inputs = torch.cat(inputs, dim=0)
        targets = torch.as_tensor(targets)
        return inputs, targets

    def __len__(self) -> int:
        """The number of samples in the dataset."""
        return self.dataset_size



import pytest

from torch_uncertainty.models.vgg.packed import packed_vgg
from torch_uncertainty.models.vgg.std import vgg


class TestVGGs:
    """Testing the VGG std class."""

    def test_main(self):
        vgg(in_channels=1, num_classes=10, arch=11, style="cifar")
        packed_vgg(
            in_channels=2,
            num_classes=10,
            arch=11,
            alpha=2,
            num_estimators=2,
            gamma=1,
        )

    def test_errors(self):
        with pytest.raises(ValueError, match="Unknown VGG arch"):
            vgg(in_channels=1, num_classes=10, arch=12, style="cifar")
        with pytest.raises(ValueError, match="Unknown VGG arch"):
            packed_vgg(
                in_channels=2,
                num_classes=10,
                arch=12,
                alpha=2,
                num_estimators=2,
                gamma=1,
            )



from functools import partial

import pytest
import torch
import torchvision.transforms as T
from torch import nn

from tests._dummies.dataset import DummyClassificationDataset
from torch_uncertainty.layers.mc_batch_norm import MCBatchNorm2d
from torch_uncertainty.models.lenet import lenet
from torch_uncertainty.post_processing import MCBatchNorm


class TestMCBatchNorm:
    """Testing the MCBatchNorm wrapper."""

    def test_main(self):
        """Test initialization."""
        mc_model = lenet(1, 1, norm=partial(MCBatchNorm2d, num_estimators=2))
        stoch_model = MCBatchNorm(mc_model, num_estimators=2, convert=False, mc_batch_size=1)

        model = lenet(1, 1, norm=nn.BatchNorm2d)
        stoch_model = MCBatchNorm(
            nn.Sequential(model),
            num_estimators=2,
            convert=True,
            mc_batch_size=1,
        )
        dataset = DummyClassificationDataset(
            "./",
            num_channels=1,
            image_size=16,
            num_classes=1,
            num_images=2,
            transform=T.ToTensor(),
        )
        stoch_model.fit(dataset=dataset)
        stoch_model.train()
        stoch_model(torch.randn(1, 1, 20, 20))
        stoch_model.eval()
        stoch_model(torch.randn(1, 1, 20, 20))

        stoch_model = MCBatchNorm(num_estimators=2, convert=False, mc_batch_size=1)
        stoch_model.set_model(mc_model)

    def test_errors(self):
        """Test errors."""
        model = nn.Identity()
        with pytest.raises(ValueError):
            MCBatchNorm(model, num_estimators=0, convert=True)
        with pytest.raises(ValueError, match="mc_batch_size must be a positive integer"):
            MCBatchNorm(model, num_estimators=1, convert=True, mc_batch_size=-1)
        with pytest.raises(ValueError):
            MCBatchNorm(model, num_estimators=1, convert=False)
        with pytest.raises(ValueError):
            MCBatchNorm(model, num_estimators=1, convert=True)
        model = lenet(1, 1, norm=nn.BatchNorm2d)
        stoch_model = MCBatchNorm(model, num_estimators=4, convert=True, mc_batch_size=1)
        dataset = DummyClassificationDataset(
            "./",
            num_channels=1,
            image_size=16,
            num_classes=1,
            num_images=2,
            transform=T.ToTensor(),
        )
        stoch_model.eval()
        with pytest.raises(RuntimeError):
            stoch_model(torch.randn(1, 1, 20, 20))

        with pytest.raises(ValueError):
            stoch_model.fit(dataset=dataset)
